from pycsp3.problems.data.parsing import *

nBoats = next_int()
data["nPeriods"] = next_int()
data["boats"] = [OrderedDict([("capacity", v2), ("crewSize", v3)]) for _ in range(nBoats) for (v1, v2, v3) in [(next_int(), next_int(), next_int())]]



from pycsp3.classes.auxiliary.ptypes import TypeVar
from pycsp3.classes.auxiliary.values import IntegerValue, IntegerInterval, SymbolicValue


class Domain:
    def __init__(self, *args):
        self.type = None
        self.original_values = []
        self._add_value(*args)
        assert self.type
        self.original_values.sort()
        self.values = []

    def set_type(self, type):
        if self.type is None:
            self.type = type
        assert self.type == type, "In a domain, values must be either all integer or all symbolic values"

    def get_type(self):
        return self.type

    def _add_value(self, arg):
        assert isinstance(arg, (tuple, list, set, range, int, str)), "Bad type for the domain " + str(arg) + " of type " + str(type(arg))
        if isinstance(arg, (tuple, list, set)):
            for a in arg:
                self._add_value(a)
        elif isinstance(arg, range):
            if arg.step == 1 and arg.stop - arg.start > 2:
                self.original_values.append(IntegerInterval(arg.start, arg.stop - 1))
                self.set_type(TypeVar.INTEGER)
            else:
                self._add_value(set(arg))
        elif isinstance(arg, int):
            self.original_values.append(IntegerValue(arg))
            self.set_type(TypeVar.INTEGER)
        elif isinstance(arg, str):
            self.original_values.append(SymbolicValue(arg))
            self.set_type(TypeVar.SYMBOLIC)

    def __iter__(self):
        return self.original_values[0].__iter__() if len(self.original_values) == 1 and isinstance(self.original_values[0],
                                                                                                   IntegerInterval) else self.original_values.__iter__()

    def __getitem__(self, item):
        return self.original_values[0].__getitem__(item) if len(self.original_values) == 1 and isinstance(self.original_values[0], IntegerInterval) \
            else self.original_values.__getitem__(item)

    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        return " ".join(str(v) for v in self.original_values)

    def smallest_value(self):
        return self.original_values[0].smallest()

    def greatest_value(self):
        return self.original_values[-1].greatest()

    def all_values(self):
        if len(self.values) == 0:
            if len(self.original_values) == 1 and isinstance(self.original_values[0], IntegerInterval):
                self.values = range(self.original_values[0].inf, self.original_values[0].sup + 1)
            else:
                self.values = sorted(v.value if isinstance(v, IntegerValue) else v for v in self)
        return self.values

    def is_binary(self):
        zero, one = False, False
        for v in self.original_values:
            if isinstance(v, IntegerInterval):
                if not v.is_binary():
                    return False
                else:
                    zero = one = True
            elif isinstance(v, IntegerValue):
                if v.value == 0:
                    zero = True
                elif v.value == 1:
                    one = True
                else:
                    return False
        return zero and one



"""
Problem 006 on CSPLib

Examples of Execution:
  python3 GolombRuler.py
  python3 GolombRuler.py -data=10
  python3 GolombRuler.py -data=10 -variant=dec
  python3 GolombRuler.py -data=10 -variant=aux
"""

from pycsp3 import *

n = data or 8
ub = n * n + 1  # a trivial upper-bound of an optimal ruler length

# x[i] is the position of the ith tick
x = VarArray(size=n, dom=range(ub))

if not variant():
    satisfy(
        # all distances are different
        AllDifferent(abs(x[i] - x[j]) for i, j in combinations(range(n), 2))
    )
elif variant("dec"):
    satisfy(
        # all distances are different
        abs(x[i] - x[j]) != abs(x[k] - x[l]) for i, j in combinations(range(n), 2) for k, l in combinations(range(i + 1, n), 2)
    )
elif variant("aux"):
    # y[i][j] is the distance between x[i] and x[j], for i strictly less than j
    y = VarArray(size=[n, n], dom=lambda i, j: range(1, ub) if i < j else None)

    satisfy(
        # all distances are different
        AllDifferent(y),

        # linking variables from both arrays
        [x[j] == x[i] + y[i][j] for i, j in combinations(range(n), 2)]
    )
    annotate(decision=x)

satisfy(
    # tag(symmetry-breaking)
    [x[0] == 0, Increasing(x, strict=True)]
)

minimize(
    # minimizing the position of the rightmost tick
    Maximum(x)
)



from pycsp3.problems.data.parsing import *

while line()[0] == '%':
    next_line()  # To pass comments

t = numbers_in(line())
assert len(t) == 3 and t[2] == 1

n = t[0]
arcs = [[0 for _ in range(n)] for _ in range(n)]
for i in range(n):
    t = numbers_in(next_line())
    for j in range(0, len(t), 2):
        arcs[i][t[j] - 1] = t[j + 1]

data["nNodes"] = n
data["arcs"] = arcs

# data.nNodes = 20
# maxWeight = 30
# p = 0.8
# data.arcs=[]
# for i in range(data.nNodes):
#    tmp = []
#    for j in range(data.nNodes):
#        if random.uniform(0,1) > p:
#           tmp.append(random.randint(0,maxWeight))
#        else:tmp.append(0)



"""
Four friends (two women named Debra and Janet, and two men named Hugh and Rick) found that each of them is allergic to something different:
eggs, mold, nuts and ragweed.
We would like to match each one's surname (Baxter, Lemon, Malone and Fleet) with his or her allergy.
We know that:
 - Rick is not allergic to mold
 - Baxter is allergic to eggs
 - Hugh is neither surnamed Lemon nor Fleet
 - Debra is allergic to ragweed
 - Janet (who isn't Lemon) is neither allergic to eggs nor to mold

Execution:
  python3 Allergy.py
"""

from pycsp3 import *

friends = Debra, Janet, Hugh, Rick = "Debra", "Janet", "Hugh", "Rick"

# foods[i] is the friend allergic to the ith food
eggs, mold, nuts, ragweed = foods = VarArray(size=4, dom=friends)

# surnames[i] is the friend with the ith surname
baxter, lemon, malone, fleet = surnames = VarArray(size=4, dom=friends)

satisfy(
    AllDifferent(foods),
    AllDifferent(surnames),

    mold != Rick,
    eggs == baxter,
    lemon != Hugh,
    fleet != Hugh,
    ragweed == Debra,
    lemon != Janet,
    eggs != Janet,
    mold != Janet
)



from pycsp3 import *
import math

maxArity, maxHeight, sentence, grammar, tokens, costs = data
nWords, nLevels, nTokens = len(sentence), len(sentence) * 2, len(tokens)
lengths = [nWords] + [nWords - math.floor((i + 1) / 2) + 1 for i in range(1, nLevels)]

# c[i][j] is the cost of the jth word at the ith level
c = VarArray(size=[nLevels, nWords], dom=lambda i, j: costs if (i == 0 or i % 2 == 1) and j < lengths[i] else {0})

# l[i][j] is the label of the jth word at the ith level
l = VarArray(size=[nLevels, nWords], dom=lambda i, j: range(nTokens + 1) if j < lengths[i] else {0})

# a[i][j] is the arity of the jth word at the ith level
a = VarArray(size=[nLevels, nWords], dom=lambda i, j: range(maxArity + 1) if i % 2 == 1 and j < lengths[i] else {0})

# x[i][j] is the index of the jth word at the ith level
x = VarArray(size=[nLevels, nWords], dom=lambda i, j: range(lengths[i]) if 0 < i and i % 2 == 0 and j < lengths[i] else {0})

s = VarArray(size=nLevels - 2, dom=lambda i: range(lengths[i + 1]))


def table_for(vector_length):
    arity = vector_length + 2
    table = {tuple([ANY] * (arity - 1) + [0])}
    for i in range(vector_length):
        for j in range(1, len(tokens) + 1):
            t = [ANY] * arity
            t[0] = i
            t[i + 1] = j
            t[arity - 1] = j
            table.add(tuple(t))
    return table


def table_for_grammar(n):
    return {tuple(ANY if v == 2147483646 else v for v in ele) for ele in grammar[n]}


def select_label(i):
    return [var for j, var in enumerate(l[i]) if j < lengths[i]]


def n_possible_sons(i, j):
    return min(lengths[i - 1] - j, maxArity)


def predicate(l, a, i, j, lengths):
    r = range(1 if i != 0 and j == lengths[i] - 1 else 0, min(j + 1, maxArity))
    return xor(*([l[i][j] == 0] + [a[i + 1][j - k] >= k + 1 for k in r]))


satisfy(
    [Count(select_label(i), value=0) == s[i - 1] for i in range(1, nLevels - 1)],

    [s[i - 1] == s[i] for i in range(1, nLevels - 1, 2)],

    # on row 0, costs are 0
    [c[0][j] == 0 for j in range(nWords)],

    # on row 0, the jth label is the jth word of the sentence
    [l[0][j] == sentence[j] for j in range(nWords)],

    # on column 0, labels are 0
    [l[i][0] > 0 for i in range(1, nLevels)],

    [a[p][0] > 0 for p in range(1, nLevels, 2)],

    [a[i][j] <= lengths[i - 1] - j for i in range(1, nLevels, 2) for j in range(1, nWords) if j < lengths[i] and j + maxArity > lengths[i - 1]],

    [(x[i][0] == 0, l[i][0] == l[i - 1][0]) for i in range(2, nLevels, 2)],

    [[x[i][j]] + select_label(i - 1) + [l[i][j]] in table_for(len(select_label(i - 1))) for i in range(2, nLevels, 2) for j in range(1, lengths[i])],

    [
        [
            x[i][j] >= j,
            imply(l[i][j] == 0, x[i][j] == lengths[i] - 1),
            imply(l[i][j] > 0, x[i][j] > x[i][j - 1]),
            imply(l[i][j - 1] == 0, l[i][j] == 0)
        ] for i in range(2, nLevels, 2) for j in range(1, lengths[i])
        ],

    # grammar a
    [
        (iff(l[i][j] == 0, a[i][j] == 0),
         [a[i][j], l[i][j]] + [l[i - 1][k] for k in range(j, j + n_possible_sons(i, j))] + [c[i][j]] in table_for_grammar(n_possible_sons(i, j)))
        for i in range(1, nLevels, 2) for j in range(lengths[i])
        ],

    [predicate(l, a, i, j, lengths) for i in range(0, nLevels, 2) for j in range(nWords) if j < lengths[i]],

    l[2 * maxHeight][1] == 0 if 0 < maxHeight and 2 * maxHeight < l.length else None

)

minimize(
    Sum(c)
)



from pycsp3.classes.entities import CtrEntities, EBlock, EGroup, ESlide, EToGather, EToSatisfy
from pycsp3.classes.main.variables import Variable


def handle_slides():
    detect_slides_recursively(CtrEntities.items)
    # canonize_blocks_of_slides(CtrEntities.items)  # TODO desactivated because very limited interest and problem with holes (see at bottom)


def detect_slides_recursively(ctr_entities):
    for ce in ctr_entities:
        if isinstance(ce, (EBlock, EToGather, EToSatisfy)):
            detect_slides_recursively(ce.entities)
        elif isinstance(ce, ESlide) and len(ce.entities) == 1:
            son = ce.entities[0]
            if isinstance(son, EToGather) and len(son.entities) == 1:
                grand_son = son.entities[0]
                if isinstance(grand_son, EGroup) and len(grand_son.entities) > 0:
                    res = _identify_slide(ce.entities[0].entities[0])
                    if res:
                        ce.scope, ce.offset, ce.circular = res  # the slide is validated: information for it is recorded
                    else:
                        print("Warning: The slide is not really a slide but a group")


def _identify_slide(group):
    def _possible_offset(all_args):  # attempt to find a simple offset
        for arg1 in all_args[1]:
            for i, arg0 in enumerate(all_args[0]):
                if arg0 == arg1:
                    return i
        return None

    if isinstance(group.abstraction, dict) and len([v for v in group.abstraction.values() if "%" in str(v)]) != 1:
        return None  # if more than one difference, we currently don't know how to slide (compute it automatically)

    all_args = group.original_all_args if hasattr(group, "original_all_args") else group.all_args

    # we compute the global scope
    t = [x for args in all_args for x in args]
    if any(not isinstance(x, Variable) for x in t):
        return None
    scope = [x for i, x in enumerate(t) if x not in t[0:i]]  # in O(n^2) but should not be a big deal here

    ''' trying to recognize a normal slide '''
    arity = len(all_args[0])
    sliding_scope = [[x for x in scope[i:i + arity]] for i in range(0, len(scope) - arity + 1, 1)]  # OFFSET 1
    if sliding_scope == all_args:
        return scope, 1, False

    ''' trying to recognize a circular Slide '''
    if sliding_scope + [[scope[i] for i in range(len(scope) - arity + 1, len(scope))] + [scope[0]]] == all_args:
        return scope, 1, True

    ''' trying to recognize a slide with an offset not equal to 1'''
    offset = _possible_offset(all_args)
    if offset is None or offset <= 1:
        return None
    sliding_scope = [[x for x in scope[i:i + arity]] for i in range(0, len(scope) - arity + 1, offset)]
    if sliding_scope == all_args:
        return scope, offset, False
    if sliding_scope + [[scope[i] for i in range(len(scope) - arity + 1, len(scope))] + [scope[0]]] == all_args:
        return scope, offset, True
    return None


# gathering slides : currently, it is not activated

def canonize_blocks_of_slides(ctr_entities, parent=None):
    for i, ce in enumerate(ctr_entities):
        if isinstance(ce, (EBlock, EToGather, EToSatisfy)):
            canonize_blocks_of_slides(ce.entities, ce)
    if not isinstance(parent, (EBlock, EToSatisfy)):
        indexes = [i for i, ce in enumerate(ctr_entities) if isinstance(ce, ESlide)]
        if len(indexes) > 1 and all(ctr_entities[indexes[0]].same_type_and_basic_attributes(ctr_entities[i]) for i in indexes):
            _build_block_of_slides(ctr_entities, indexes)


def _build_block_of_slides(ctr_entities, indexes):
    block = EBlock([ctr_entities[i] for i in indexes])
    block.copy_basic_attributes_of(ctr_entities[indexes[0]])
    for i in indexes:
        ctr_entities[i].clear_basic_attributes()
        ctr_entities[i] = None
    # CtrEntities.remove(block)  # todo can we do that?
    ctr_entities[indexes[0]] = block
    # TODO we should not keep holes (None) in arrays as above



"""
Problem 061 on CSPLib

Example of Execution:
  python3 Rcpsp.py -data=Rcpsp_j30-01-01.json
"""

from pycsp3 import *

horizon, capacities, jobs = data
nJobs = len(jobs)


def cumulative_for(k):
    origins, lengths, heights = zip(*[(s[i], duration, quantities[k]) for i, (duration, _, quantities) in enumerate(jobs) if quantities[k] > 0])
    return Cumulative(origins=origins, lengths=lengths, heights=heights)


# s[i] is the starting time of the ith job
s = VarArray(size=nJobs, dom=lambda i: {0} if i == 0 else range(horizon))

satisfy(
    # precedence constraints
    [s[i] + duration <= s[j] for i, (duration, successors, _) in enumerate(jobs) for j in successors],

    # resource constraints
    [cumulative_for(k) <= capacity for k, capacity in enumerate(capacities)]
)

minimize(
    s[- 1]
)


# Note that:

# a) using zip is compacter than writing something like:
#   indexes = [i for i in range(nJobs) if jobs[i].requiredQuantities[k] > 0]
#   origins = [s[i] for i in indexes]
#   lengths = [jobs[i].duration for i in indexes]
#   heights = [jobs[i].requiredQuantities[k] for i in indexes]

# b) using namedtuple facility allows us to get directly fields of the tuples. Instead, we could write:
#   origins, lengths, heights = zip(*[(s[i], job.duration, job.requiredQuantities[k]) for i, job in enumerate(jobs) if job.requiredQuantities[k] > 0])



"""
See http://www.schedulingbenchmarks.org/nurseinstances1_24.html

Example of Execution:
  python3 NurseRostering.py -data=NurseRostering_00.json
"""

from pycsp3 import *

nDays, shifts, staffs, covers = data
if shifts[-1].id != "_off":  # if not present, we add first a dummy 'off' shift (a named tuple of the right class)
    shifts.append(shifts[0].__class__("_off", 0, None))
off = len(shifts) - 1  # value for _off
nWeeks, nShifts, nStaffs = nDays // 7, len(shifts), len(staffs)

on_r = [[next((r for r in staff.onRequests if r.day == day), None) if staff.onRequests else None for day in range(nDays)] for staff in staffs]
off_r = [[next((r for r in staff.offRequests if r.day == day), None) if staff.offRequests else None for day in range(nDays)] for staff in staffs]
lengths = [shift.length for shift in shifts]

kmax = [staff.maxConsecutiveShifts for staff in staffs]
kmin = [staff.minConsecutiveShifts for staff in staffs]
kday = [staff.minConsecutiveDaysOff for staff in staffs]

sp = {shifts[i].id: i for i in range(nShifts)}  # position of shifts in the list 'shifts'
table = {(sp[s1.id], sp[s2]) for s1 in shifts if s1.forbiddenFollowingShifts for s2 in s1.forbiddenFollowingShifts}  # rotation


def costs(day, shift):
    if shift == off:
        return [0] * (nStaffs + 1)
    r, wu, wo = covers[day][shift]
    return [abs(r - i) * (wu if i <= r else wo) for i in range(nStaffs + 1)]


def automaton_min_consecutive(k, for_shifts):
    q = Automaton.q  # for building state names
    range_off = range(nShifts - 1, nShifts)  # a range with only one value (off)
    range_others = range(nShifts - 1)  # a range with all other values
    r1, r2 = range_off if for_shifts else range_others, range_others if for_shifts else range_off
    t = ([(q(0), r1, q(1)), (q(0), r2, q(k + 1)), (q(1), r1, q(k + 1))]
         + [(q(i), r2, q(i + 1)) for i in range(1, k + 1)]
         + [(q(k + 1), range(nShifts), q(k + 1))])
    return Automaton(start=q(0), final=q(k + 1), transitions=t)


# x[d][p] is the shift at day d for person p (value 'off' denotes off)
x = VarArray(size=[nDays, nStaffs], dom=range(nShifts))

# nd[p][s] is the number of days such that person p works with shift s
nd = VarArray(size=[nStaffs, nShifts], dom=lambda p, s: range((nDays if s == off else staffs[p].maxShifts[s]) + 1))

# np[d][s] is the number of persons working on day d with shift s
np = VarArray(size=[nDays, nShifts], dom=range(nStaffs + 1))

# wk[p][w] is 1 iff the week-end w is worked by person p
wk = VarArray(size=[nStaffs, nWeeks], dom={0, 1})

# cn[p][d] is the cost of not satisfying the on-request (if it exists) of person p on day d
cn = VarArray(size=[nStaffs, nDays], dom=lambda p, d: {0, on_r[p][d].weight} if on_r[p][d] else None)

# cf[p][d] is the cost of not satisfying the off-request (if it exists) of person p on day d
cf = VarArray(size=[nStaffs, nDays], dom=lambda p, d: {0, off_r[p][d].weight} if off_r[p][d] else None)

# cc[d][s] is the cost of not satisfying cover for shift s on day d
cc = VarArray(size=[nDays, nShifts], dom=lambda d, s: costs(d, s))

satisfy(
    # days off for staff
    [x[d][p] == off for d in range(nDays) for p in range(nStaffs) if d in staffs[p].daysOff],

    # computing number of days
    [Count(x[:, p], value=s) == nd[p][s] for p in range(nStaffs) for s in range(nShifts)],

    # computing number of persons
    [Count(x[d], value=s) == np[d][s] for d in range(nDays) for s in range(nShifts)],

    # computing worked week-ends
    [(imply(x[w * 7 + 5][p] != off, wk[p][w]), imply(x[w * 7 + 6][p] != off, wk[p][w])) for p in range(nStaffs) for w in range(nWeeks)],

    # rotation shifts
    [Slide((x[i][p], x[i + 1][p]) not in table for i in range(nDays - 1)) for p in range(nStaffs)] if len(table) > 0 else None,

    # maximum number of worked week-ends
    [Sum(wk[p]) <= staffs[p].maxWeekends for p in range(nStaffs)],

    # minimum and maximum number of total worked minutes
    [nd[p] * lengths in range(staff.minTotalMinutes, staff.maxTotalMinutes + 1) for p, staff in enumerate(staffs)],

    # maximum consecutive worked shifts
    [Count(x[i:i + kmax[p] + 1, p], value=off) >= 1 for p in range(nStaffs) for i in range(nDays - kmax[p])],

    # minimum consecutive worked shifts
    [x[i: i + kmin[p] + 1, p] in automaton_min_consecutive(kmin[p], True) for p in range(nStaffs) for i in range(nDays - kmin[p])],

    # managing off days on schedule ends
    [(imply(x[0][p] != off, x[i][p] != off), imply(x[- 1][p] != off, x[- 1 - i][p] != off)) for p in range(nStaffs) if kmin[p] > 1 for i in range(1, kmin[p])],

    # minimum consecutive days off
    [x[i: i + kday[p] + 1, p] in automaton_min_consecutive(kday[p], False) for p in range(nStaffs) for i in range(nDays - kday[p])],

    # cost of not satisfying on requests
    [iff(x[d][p] == sp[on_r[p][d].shift], cn[p][d] == 0) for p in range(nStaffs) for d in range(nDays) if on_r[p][d]],

    # cost of not satisfying off requests 
    [iff(x[d][p] == sp[off_r[p][d].shift], cf[p][d] != 0) for p in range(nStaffs) for d in range(nDays) if off_r[p][d]],

    # cost of under or over covering 
    [(np[d][s], cc[d][s]) in enumerate(costs(d, s)) for d in range(nDays) for s in range(nShifts)]
)

minimize(
    Sum(cn) + Sum(cf) + Sum(cc)
)


# [iff(x[d][p] == s, cn[p][d] == 0) for (p,d,s) in [(p,d, sp[on_r[p][d].shift]) for p in range(nStaffs) for d in range(nDays) if on_r[p][d]]],



"""
Problem 029 on CSPLib

Examples of Execution:
  python3 QueenAttacking.py
  python3 QueenAttacking.py -data=6
  python3 QueenAttacking.py -data=6 -variant=aux
  python3 QueenAttacking.py -data=6 -variant=hybrid
  python3 QueenAttacking.py -data=6 -variant=table
"""

from pycsp3 import *

n = data or 8
primes = all_primes(n * n)
m = len(primes)


def row(var):
    return var // n


def col(var):
    return var % n


# q is the cell for the queen
q = Var(dom=range(n * n))

# x[i] is the cell for the i+1th value
x = VarArray(size=n * n, dom=range(n * n))

if not variant():
    satisfy(
        # all values are put in different cells
        AllDifferent(x),

        # ensuring a knight move between two successive values
        [(d1 == 1) & (d2 == 2) | (d1 == 2) & (d2 == 1)
         for d1, d2 in [(abs(row(x[i]) - row(x[i + 1])), abs(col(x[i]) - col(x[i + 1]))) for i in range(n * n - 1)]]
    )

    minimize(
        # minimizing the number of free primes
        Sum((q == x[j]) | (row(q) != row(x[j])) & (col(q) != col(x[j])) & (abs(row(q) - row(x[j])) != abs(col(q) - col(x[j]))) for j in [p - 1 for p in primes])
    )

elif variant("aux"):
    # p[j] is 1 iff the j+1th prime number is not attacked by a queen
    p = VarArray(size=m, dom={0, 1})

    satisfy(
        # all values are put in different cells
        AllDifferent(x),

        # ensuring a knight move between two successive values
        [(d1 == 1) & (d2 == 2) | (d1 == 2) & (d2 == 1)
         for d1, d2 in [(abs(row(x[i]) - row(x[i + 1])), abs(col(x[i]) - col(x[i + 1]))) for i in range(n * n - 1)]],

        # determining if prime numbers are attacked by the queen
        [iff(p[j], (q == x[k]) | (row(q) != row(x[k])) & (col(q) != col(x[k])) & (abs(row(q) - row(x[k])) != abs(col(q) - col(x[k]))))
         for j, k in enumerate(p - 1 for p in primes)]
    )

    minimize(
        # minimizing the number of free primes
        Sum(p)
    )

elif variant("hybrid"):  # hybrid as compilation will build and combine both intension and extension constraints for the list of constraints about knight moves
    satisfy(
        # all values are put in different cells
        AllDifferent(x),

        # ensuring a knight move between two successive values
        [(abs(row(x[i]) - row(x[i + 1])), abs(col(x[i]) - col(x[i + 1]))) in {(1, 2), (2, 1)} for i in range(n * n - 1)]
    )

    minimize(
        # minimizing the number of free primes
        Sum((q == x[j]) | (row(q) != row(x[j])) & (col(q) != col(x[j])) & (abs(row(q) - row(x[j])) != abs(col(q) - col(x[j]))) for j in [p - 1 for p in primes])
    )

elif variant("table"):
    def neighbours(r1, c1):
        return [(r1 + r2) * n + c1 + c2 for (r2, c2) in [(-2, -1), (-2, 1), (-1, -2), (-1, 2), (1, -2), (1, 2), (2, -1), (2, 1)] if
                0 <= r1 + r2 < n and 0 <= c1 + c2 < n]


    table1 = {(i, j) for i in range(n * n) for j in neighbours(i // n, i % n)}
    table2 = {(i, j, 1 if (i == j) | (i // n != j // n) & (i % n != j % n) & (abs(i // n - j // n) != abs(i % n - j % n)) else 0) for i in range(n * n) for j in
              range(n * n)}

    # p[j] is 1 iff the j+1th prime number is not attacked by a queen
    p = VarArray(size=m, dom={0, 1})

    satisfy(
        # all values are put in different cells
        AllDifferent(x),

        # ensuring a knight move between two successive values
        [(x[i], x[i + 1]) in table1 for i in range(n * n - 1)],

        # determining if prime numbers are attacked by the queen
        [(q, x[k], p[j]) in table2 for j, k in enumerate(p - 1 for p in primes)]
    )

    minimize(
        # minimizing the number of free primes
        Sum(p)
        # (q == x[j]) | (row(q) != row(x[j])) & (col(q) != col(x[j])) & (abs(row(q) - row(x[j])) != abs(col(q) - col(x[j]))) for j in [p - 1 for p in primes])
    )



# Note that:

# a) the variant hybrid involves the automatic introduction of auxiliary variables



from pycsp3.problems.tests.tester import Tester, run

NAME = "g2_academic"

diff = (Tester(NAME)
        .add("CryptoPuzzle", data="[DONALD,GERALD,ROBERT]")  # ok, aggregation more compact (subgroups in JvCSP)
        .add("DeBruijnSequence", data="[2,5]")
        .add("DiamondFree", data="10")
        .add("GolombRuler", data="8", variant="dec")  # ok, predicate expressions more general (2 sub-groups in JvCSP)
        .add("LowAutocorrelation", data="16")
        .add("Ortholatin", data="10")  # ok, instantiation more compact
        .add("SocialGolfers", data="[4,4,5]")  # ok, instantiation more compact
        .add("SocialGolfers", data="[4,4,5]", variant="01")  # ok, block not formed in JvCSP for lex constraints
        .add("SportsScheduling", data="10")  # ok, class 'dummy-week' put before if in JvCSP
        .add("StillLife", data="[8,8]")
        .add("StillLife", data="[8,8]", variant="wastage")
        )

same = (Tester(NAME)
        .add("AllInterval", data="10")
        .add("AllInterval", data="10", variant="aux")
        .add("Bibd", data="[9,0,0,3,9]")
        .add("Bibd", data="[9,0,0,3,9]", variant="aux")
        .add("BoardColoration", data="[8,10]")
        .add("ChangeMaking", data="10")
        .add("ColouredQueens", data="10")
        .add("CoveringArray", data="[3,5,2,10]")
        .add("CryptoPuzzle", data="[SEND,MORE,MONEY]", variant="carry")
        .add("Dubois", data="10")
        .add("Enigme5", data="[364,55440]")
        .add("GolombRuler", data="8")
        .add("GolombRuler", data="8", variant="aux")
        .add("GracefulGraph", data="[3,5]")
        .add("KnightTour", data="16")
        .add("KnightTour", data="16", variant="table-2")
        .add("KnightTour", data="16", variant="table-3")
        .add("Langford", data="[16,4]")
        .add("LangfordBin", data="16")
        .add("MagicHexagon", data="[4,10]")
        .add("MagicSequence", data="10")
        .add("NumberPartitioning", data="10")
        .add("Opd", data="[4,4,4]")
        .add("Opd", data="[4,6,4]", variant="aux")
        .add("PeacableArmies", data="10", variant="m1")
        .add("PeacableArmies", data="10", variant="m2")
        .add("QuasiGroup", data="10", variant="aux-v3")
        .add("QuasiGroup", data="10", variant="aux-v4")
        .add("QuasiGroup", data="10", variant="aux-v5")
        .add("QuasiGroup", data="10", variant="aux-v7")
        .add("QueenAttacking", data="10")
        .add("Queens", data="10")
        .add("Queens", data="10", variant="v1")
        .add("Queens", data="10", variant="v2")
        .add("Ramsey", data="10")
        .add("SchurrLemma", data="[10,10]")
        .add("SchurrLemma", data="[10,10]", variant="mod")
        .add("Steiner3", data="6")
        .add("Talisman", data="[4,4]")
        .add("WaterBucket", data="[8,5,3,4,4,0,8]")
        )

xcsp = (Tester(NAME)
        .add("AllInterval", data="10")  # 148 solutions (296 without symmetry-breaking)
        .add("AllInterval", data="10", variant="aux")  # 104 solutions  (296 without symmetry-breaking)
        .add("Bibd", data="[6,0,0,3,8]")  # 494 solutions
        .add("Bibd", data="[6,0,0,3,8]", variant="aux")  # 494 solutions
        .add("BoardColoration", data="[8,10]")  # optimum 2
        .add("ChangeMaking", data="10")  # optimum 1
        .add("ChangeMaking", data="10", variant="compact")  # optimum 1
        .add("CoinsGrid", data="[10,4]")  # optimum 98
        .add("ColouredQueens", data="6")  # unsat
        .add("CostasArray", data="10")  # 2160 solutions
        .add("CoveringArray", data="[3,5,2,10]")  # sat
        .add("CryptoPuzzle", data="[SEND,MORE,MONEY]")  # 1 solution
        .add("CryptoPuzzle", data="[SEND,MORE,MONEY]", variant="carry")  # 1 solution
        .add("CryptoPuzzle", data="[DONALD,GERALD,ROBERT]")  # 1 solution
        .add("DeBruijnSequence", data="[2,5]")  # 2048 solutions
        .add("DiamondFree", data="8")  # 17 solutions
        .add("Dubois", data="10")  # unsat
        .add("Enigme5", data="[364,55440]")
        .add("GolombRuler", data="8")  # optimum 34
        .add("GolombRuler", data="8", variant="dec")  # optimum 34
        .add("GolombRuler", data="8", variant="aux")  # optimum 34
        .add("GracefulGraph", data="[3,5]")  # sat
        .add("KnightTour", data="16", options_py="-recognizeSlides")
        .add("KnightTour", data="16", variant="table-2", options_py="-recognizeSlides")
        .add("KnightTour", data="16", variant="table-3", options_py="-recognizeSlides")
        .add("Langford", data="[3,10]")  # 10 solutions
        .add("LangfordBin", data="8")  # 300 solutions
        .add("LowAutocorrelation", data="16")  # optimum 24
        .add("MagicHexagon", data="[4,3]")  # sat
        .add("MagicSequence", data="10")  # 1 solution
        .add("NumberPartitioning", data="8")  # 1 solution
        .add("NumberPartitioning", data="10")  # unsat
        .add("Opd", data="[4,4,4]")  # optimum 4
        .add("Opd", data="[4,6,4]", variant="aux")  # optimum 3
        .add("Ortholatin", data="5")  # 4 solutions
        .add("PeacableArmies", data="6", variant="m1")  # optimum 5
        .add("PeacableArmies", data="6", variant="m2")  # optimum 5
        .add("QuasiGroup", data="8", variant="base-v3")  # 12960 solutions
        .add("QuasiGroup", data="5", variant="base-v4")  # 12 solutions
        .add("QuasiGroup", data="8", variant="base-v5")  # 720 sols
        .add("QuasiGroup", data="8", variant="base-v6")  # 1440 sols
        .add("QuasiGroup", data="9", variant="base-v7")  # 5040 sols
        .add("QuasiGroup", data="8", variant="aux-v3")  # 12960 solutions
        .add("QuasiGroup", data="5", variant="aux-v4")  # 12 solutions
        .add("QuasiGroup", data="8", variant="aux-v5")  # 720 sols
        .add("QuasiGroup", data="9", variant="aux-v7")  # 5040 sols
        .add("QueenAttacking", data="6", options_py="-recognizeSlides")  # optimum 0
        .add("QueenAttacking", data="6", variant="aux", options_py="-recognizeSlides")  # optimum 0
        .add("QueenAttacking", data="6", variant="hybrid", options_py="-recognizeSlides")  # optimum 0
        .add("QueenAttacking", data="6", variant="table", options_py="-recognizeSlides")  # optimum 0
        .add("Queens", data="10")  # 724 solutions
        .add("Queens", data="10", variant="v1")  # 724 solutions
        .add("Queens", data="10", variant="v2")  # 724 solutions
        .add("Ramsey", data="10")  # optimum 2
        .add("SchurrLemma", data="[6,6]")  # 39870 solutions
        .add("SchurrLemma", data="[8,8]", variant="mod")  # 141120 solutions
        .add("SocialGolfers", data="[4,4,5]")  # 2 solutions
        .add("SocialGolfers", data="[4,4,5]", variant="01")  # 2 solutions
        .add("SportsScheduling", data="6")  # 10 solutions
        .add("SportsScheduling", data="6", variant="dummy")  # 10 solutions
        .add("Steiner3", data="7")  # 151200 solutions
        .add("StillLife", data="[7,7]")  # optimum 28
        .add("StillLife", data="[8,8]", variant="wastage")  # optimum 36
        .add("Talisman", data="[4,2]")  # 34714 solutions
        .add("WaterBucket", data="[8,5,3,4,4,0,8]")  # optimum 7
        )

# .add("QuasiGroup", data="10", variant="base-v6")


run(xcsp, diff, same)



"""
Someone in the university ate Alice’s sandwich at the cafeteria. We want to find out who the culprit is.
The witnesses are unanimous about the following facts:
 - Three persons were in the cafeteria at the time of the crime: Alice, Bob, and Sascha.
 - The culprit likes Alice.
 - The culprit is taller than Alice.
 - Nobody is taller than himself.
 - If A is taller than B, then B is not taller than A.
 - Bob likes no one that Alice likes.
 - Alice likes everybody except Bob.
 - Sascha likes everyone that Alice likes.
 - Nobody likes everyone.

Execution:
  python3 Sandwich.py
"""

from pycsp3 import *

alice, bob, sascha = persons = 0, 1, 2

# culprit is among alice (0), bob (1) and sascha (2)
culprit = Var(persons)

# liking[i][j] is 1 iff the ith guy likes the jth guy
liking = VarArray(size=[3, 3], dom={0, 1})

# taller[i][j] is 1 iff the ith guy is taller than the jth guy
taller = VarArray(size=[3, 3], dom={0, 1})

satisfy(
    # the culprit likes Alice
    liking[culprit][alice] == 1,

    # the culprit is taller than Alice
    taller[culprit][alice] == 1,

    # nobody is taller than himself
    [taller[p][p] == 0 for p in persons],

    # the ith guy is taller than the jth guy iff the reverse is not true
    [taller[p1][p2] != taller[p2][p1] for p1 in persons for p2 in persons if p1 != p2],

    # Bob likes no one that Alice likes
    [imply(liking[alice][p], ~liking[bob][p]) for p in persons],

    # Alice likes everybody except Bob
    [liking[alice][p] == 1 for p in persons if p != bob],

    # Sascha likes everyone that Alice likes
    [imply(liking[alice][p], liking[sascha][p]) for p in persons],

    # nobody likes everyone
    [Count(liking[p], value=0) >= 1 for p in persons]
)



"""
Given a edge-weighted directed graph with possibly many cycles, the task is to find an acyclic sub-graph of maximal weight.

Examples of Execution:
  python3 GraphMaxAcyclic.py -data=GraphMaxAcyclic_example.json
  python3 GraphMaxAcyclic.py -data=GraphMaxAcyclic_example.json -variant=cnt
  python3 GraphMaxAcyclic.py -data=GraphMaxAcyclic_example.txt -dataparser=GraphMaxAcyclic_Parser.py
"""

from pycsp3 import *

n, arcs = data

valid_arcs = [(i, j) for i in range(n) for j in range(n) if i != j and arcs[i][j] != 0]
valid_numbers = [len([(i, j) for i in range(n) if (i, j) in valid_arcs]) for j in range(n)]

# x[i] is the number associated with the ith node; arcs are only possible from greater to lower numbers (nodes)
x = VarArray(size=n, dom=range(n))

# a[i][j] is 1 iff the arc from i to j is selected
a = VarArray(size=[n, n], dom=lambda i, j: {0, 1} if (i, j) in valid_arcs else None)

satisfy(
    # different numbers must be associated to nodes
    AllDifferent(x)
)

if not variant():
    satisfy(
        # ensuring acyclicity
        iff(x[i] > x[j], a[i][j] == 1) for (i, j) in valid_arcs
    )

elif variant("cnt"):
    satisfy(
        # ensuring acyclicity
        [imply(x[i] <= x[j], a[i][j] == 0) for (i, j) in valid_arcs],

        [Count(a[:, j], value=1) <= 3 for j in range(n) if valid_numbers[j] > 3]
    )

maximize(
    # maximising the summed weight of selected arcs
    Sum(a[i][j] * arcs[i][j] for (i, j) in valid_arcs)
)


# if modelVariant("smart"):
# c[i][j] is the cost of the link between i and j (whatever the direction)
#    c = varArray(size=[n, n], dom=lambda i, j: {arcs[i][j], arcs[j][i]}, when=lambda i, j: (arcs[i][j] != 0 or arcs[j][i] != 0) and i < j)



import inspect
import types
from collections import namedtuple
from itertools import combinations, product, permutations

from pycsp3.classes.auxiliary.conditions import Condition, ConditionValue, ne
from pycsp3.classes.auxiliary.structures import Automaton, MDD
from pycsp3.classes.auxiliary.ptypes import TypeOrderedOperator, TypeConditionOperator, TypeVar, TypeCtr, TypeCtrArg, TypeRank
from pycsp3.classes.entities import (
    EVar, EVarArray, ECtr, ECtrs, EToGather, EToSatisfy, EBlock, ESlide, EIfThenElse, EObjective, EAnnotation, AnnEntities, TypeNode, Node)
from pycsp3.classes.main.annotations import (
    AnnotationDecision, AnnotationOutput, AnnotationVarHeuristic, AnnotationValHeuristic, AnnotationFiltering, AnnotationPrepro, AnnotationSearch,
    AnnotationRestarts)
from pycsp3.classes.main.constraints import (
    ConstraintIntension, ConstraintExtension, ConstraintRegular, ConstraintMdd, ConstraintAllDifferent,
    ConstraintAllDifferentList, ConstraintAllDifferentMatrix, ConstraintAllEqual, ConstraintOrdered, ConstraintLex, ConstraintLexMatrix, ConstraintSum,
    ConstraintCount, ConstraintNValues, ConstraintCardinality, ConstraintMaximum, ConstraintMinimum, ConstraintChannel, ConstraintNoOverlap,
    ConstraintCumulative, ConstraintCircuit, ConstraintClause, PartialConstraint, ScalarProduct, auxiliary)
from pycsp3.classes.main.domains import Domain
from pycsp3.classes.main.objectives import ObjectiveExpression, ObjectivePartial
from pycsp3.classes.main.variables import Variable, VariableInteger, VariableSymbolic, NotVariable, NegVariable
from pycsp3.dashboard import options
from pycsp3.tools import curser
from pycsp3.tools.curser import OpOverrider, ListInt, ListVar, columns
from pycsp3.tools.inspector import checkType, extract_declaration_for, comment_and_tags_of, comments_and_tags_of_parameters_of
from pycsp3.tools.utilities import ANY, flatten, is_containing, is_1d_list, is_1d_tuple, is_matrix, is_square_matrix, transpose, alphabet_positions, all_primes, \
    integer_scaling
from pycsp3.compiler import default_data

''' Global Variables '''

absPython, maxPython, minPython, combinationsPython = abs, max, min, combinations


def combinations(n, size):
    return combinationsPython(n, size) if not isinstance(n, int) else combinationsPython(range(n), size)


def protect():
    return OpOverrider.disable()


''' Checking Model Variants '''


def variant(name=None):
    """"
       test
    """
    pos = -1 if not options.variant else options.variant.find("-")  # position of dash ('-') in options.variant
    if not name:
        return options.variant[0:pos] if pos != -1 else options.variant
    return options.variant[0:pos] == name if pos != -1 else options.variant == name


def subvariant(name=None):
    pos = -1 if not options.variant else options.variant.find("-")  # position of dash ('-') in options.variant
    if not name:
        return None if pos == -1 else options.variant[pos + 1:]
    return pos != -1 and options.variant[pos + 1:] == name


''' Declaring stand-alone variables and arrays '''


def Var(term=None, *others, dom=None):
    assert (term is None) != (dom is None)
    if term is not None:
        dom = flatten(term, others)
    if not isinstance(dom, Domain):
        dom = Domain(dom)
    name = extract_declaration_for("Var")
    comment, tags = comment_and_tags_of(function_name="Var")

    assert isinstance(comment, (str, type(None))), "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
    assert name not in Variable.name2obj, "The identifier " + name + " is used twice. This is not possible"
    assert dom.get_type() in {TypeVar.INTEGER, TypeVar.SYMBOLIC}, "Currently, only integer and symbolic variables are supported. Problem with " + str(dom)
    assert str(name) not in Variable.name2obj, "The identifier " + name + " is used twice. This is not possible"

    var_object = VariableInteger(name, dom) if dom.get_type() == TypeVar.INTEGER else VariableSymbolic(name, dom)
    Variable.name2obj[name] = var_object
    EVar(var_object, comment, tags)  # object wrapping the variable x
    return var_object


def VarArray(*, size, dom, comment=None):
    size = [size] if isinstance(size, int) else size
    assert all(dimension != 0 for dimension in size), "No dimension must not be equal to 0"
    checkType(size, [int])
    checkType(dom, (range, Domain, [int, range, str, Domain, type(None)], type(lambda: 0)))
    name = extract_declaration_for("VarArray")
    if comment is None and not isinstance(name, list):
        comment, tags = comment_and_tags_of(function_name="VarArray")

    assert isinstance(comment, (str, type(None))), "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
    assert not isinstance(dom, type(lambda: 0)) or len(size) == len(inspect.signature(dom).parameters), \
        "The number of arguments of the lambda must be equal to the number of dimensions of the multidimensional array"
    assert isinstance(comment, (str, type(None))), "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
    assert str(name) not in Variable.name2obj, "The identifier " + name + " is used twice. This is not possible"

    var_objects = Variable.build_variables_array(name, size, dom)
    if isinstance(name, list):
        for variable in var_objects:
            EVar(variable, None, None)  # object wrapping the variables
        return tuple(var_objects)
    else:
        EVarArray(var_objects, name, comment, tags)  # object wrapping the array of variables
        return ListVar(var_objects)


''' Posting constraints (through satisfy()) '''


def _bool_interpretation_for_in(left_operand, right_operand, bool_value):
    assert type(bool_value) is bool
    if isinstance(left_operand, Variable) and isinstance(right_operand, (tuple, list, set, frozenset, range)) and len(right_operand) == 0:
        return None
    if isinstance(left_operand, (Variable, int, str)) and isinstance(right_operand, (set, frozenset, range)):
        # it is a unary constraint of the form x in/not in set/range
        ctr = Intension(Node.build(TypeNode.IN, left_operand, right_operand) if bool_value else Node.build(TypeNode.NOTIN, left_operand, right_operand))
    elif isinstance(left_operand, PartialConstraint):  # it is a partial form of constraint (sum, count, maximum, ...)
        ctr = ECtr(left_operand.constraint.set_condition(TypeConditionOperator.IN if bool_value else TypeConditionOperator.NOTIN, right_operand))
    elif isinstance(right_operand, Automaton):  #  it is a regular constraint
        ctr = Regular(scope=left_operand, automaton=right_operand)
    elif isinstance(right_operand, MDD):  # it is a MDD constraint
        ctr = Mdd(scope=left_operand, mdd=right_operand)
    elif isinstance(left_operand, int) and (is_1d_list(right_operand, Variable) or is_1d_tuple(right_operand, Variable)):
        ctr = Count(right_operand, value=left_operand, condition=(TypeConditionOperator.GE, 1))  # atLeast1 TODO to be replaced by a member/element constraint ?
    else:  #  It is a table constraint
        if not hasattr(left_operand, '__iter__'):
            left_operand = [left_operand]
        if not bool_value and len(right_operand) == 0:
            return None
        # TODO what to do if the table is empty and bool_value is true? an error message ?
        ctr = Extension(scope=list(left_operand), table=right_operand, positive=bool_value)
    return ctr


def _complete_partial_forms_of_constraints(entities):
    for i, c in enumerate(entities):
        if isinstance(c, bool):
            assert len(curser.queue_in) > 0, "A boolean that does not represent a constraint is in the list of constraints in satisfy(): " + str(entities)
            right_operand, left_operand = curser.queue_in.popleft()
            entities[i] = _bool_interpretation_for_in(left_operand, right_operand, c)
        elif isinstance(c, ESlide):
            for ent in c.entities:
                _complete_partial_forms_of_constraints(ent.entities)
    return entities


def _wrap_intension_constraints(entities):
    for i, c in enumerate(entities):
        if isinstance(c, Node):
            entities[i] = Intension(c)  # the node is wrapped by a Constraint object (Intension)
    return entities


def IfThenElse(*args):
    return EIfThenElse(_wrap_intension_constraints(_complete_partial_forms_of_constraints(flatten(*args))))


def Slide(*args):
    entities = _wrap_intension_constraints(
        flatten(*args))  # we cannot directly complete partial forms (because it is executed before the analysis of the parameters of satisfy
    checkType(entities, [ECtr, bool])
    return ESlide([EToGather(entities)])


def satisfy(*args):
    global no_parameter_satisfy, nb_parameter_satisfy

    def _group(*_args):
        entities = _wrap_intension_constraints(_complete_partial_forms_of_constraints(flatten(*_args)))
        checkType(entities, [ECtr, ECtrs])
        return EToGather(entities)

    def _block(*_args):
        def _reorder(_entities):
            reordered_entities = []
            g = []
            for c in _entities:
                if isinstance(c, ECtr):
                    g.append(c)
                else:
                    if len(g) != 0:
                        reordered_entities.append(_group(g))
                    g.clear()
                    reordered_entities.append(c)
            if len(g) != 0:
                reordered_entities.append(_group(g))
            return reordered_entities

        entities = _wrap_intension_constraints(_complete_partial_forms_of_constraints(flatten(*_args)))
        checkType(entities, [ECtr, ECtrs])
        return EBlock(_reorder(entities))

    def _reorder(l):  # if constraints are given in (sub-)lists inside tuples; we flatten and reorder them to hopefully improve compactness
        d = dict()
        for tp in l:
            if isinstance(tp, tuple):
                for i, v in enumerate(tp):
                    d.setdefault(i, []).append(v)
            else:
                d.setdefault(0, []).append(tp)
        r = []
        for i in range(len(d)):
            r.extend(d[i])
        return r

    no_parameter_satisfy = 0
    nb_parameter_satisfy = len(args)
    comments1, comments2, tags1, tags2 = comments_and_tags_of_parameters_of(function_name="satisfy", args=args)
    t = []
    for i, arg in enumerate(args):
        if isinstance(arg, list) and len(arg) > 0:
            if isinstance(arg[0], tuple):
                arg = _reorder(arg)
            elif isinstance(arg[0], list):  # do not work if the constraints involve the operator 'in'
                for j, l in enumerate(arg):
                    if isinstance(l, list) and len(l) > 0 and isinstance(l[0], tuple):
                        arg[j] = _reorder(l)
        no_parameter_satisfy = i
        assert isinstance(arg, (ECtr, ESlide, Node, bool, list, tuple, type(None), types.GeneratorType)), "non authorized type " + str(type(arg))
        if arg is None:
            continue
        arg = list(arg) if isinstance(arg, types.GeneratorType) else arg
        comment_at_2 = any(comment != '' for comment in comments2[i])
        if isinstance(arg, (ECtr, ESlide)):  # case: simple constraint or slide
            to_post = _complete_partial_forms_of_constraints([arg])[0]
        elif isinstance(arg, Node):  # a predicate to be wrapped by a constraint (intension)
            to_post = Intension(arg)
        elif isinstance(arg, bool):  # a Boolean representing the case of a partial constraint or a node with operator in {IN, NOT IN}
            assert curser.queue_in, "An argument of satisfy() before position " + str(i) + " is badly formed"
            # assert queue_in, "A boolean that do not represents a constraint is in the list of constraints in satisfy(): " \
            #                 + str(args) + " " + str(i) + ".\nA constraint is certainly badly formed"
            other, partial = curser.queue_in.popleft()
            to_post = _bool_interpretation_for_in(partial, other, arg)
        elif any(isinstance(ele, ESlide) for ele in arg):  #  Case: Slide
            to_post = _block(arg)
        elif comment_at_2:  # Case: block
            for j, ele in enumerate(arg):
                if isinstance(arg[j], (ECtr, ESlide)):
                    arg[j].note(comments2[i][j]).tag(tags2[i][j])
                elif comments2[i][j] or tags2[i][j]:
                    arg[j] = _group(arg[j]).note(comments2[i][j]).tag(tags2[i][j])
            to_post = _block(arg)
        else:  #  Group
            to_post = _group(arg)
        if to_post:
            t.append(to_post.note(comments1[i]).tag(tags1[i]))
            # if isinstance(to_post, ESlide) and len(to_post.entities) == 1:
            #     to_post.entities[0].note(comments1[i]).tag(tags1[i])
    t.append(_group(pc == var for (pc, var) in auxiliary().collected()))
    return EToSatisfy(t)


''' Generic Constraints (intension, extension) '''


# def is_smart(table):
#     for t in table:
#         for v in t:
#             if isinstance(v, ConditionValue):
#                 return True
#     return False


def Extension(*, scope, table, positive=True):
    scope = flatten(scope)
    checkType(scope, [Variable])
    assert isinstance(table, list)
    if len(scope) > 1 and any(isinstance(v, ConditionValue) for t in table for v in t):  # if smart table
        table = sorted(list(to_ordinary_table(table, [x.dom for x in scope], keep_any=True)))

    checkType(table, [str, int, float])
    checkType(positive, bool)
    assert isinstance(table, list) and len(table) > 0, "A table must be a non-empty list of tuples or integers (or symbols)"
    assert isinstance(table[0], (tuple, int, str)), "Elements of tables are tuples or integers (or symbols)"
    # print(table)
    assert isinstance(table[0], (int, str)) or len(scope) == len(table[0]), (
        "The length of each tuple must be the same as the arity." + "Maybe a problem with slicing: you must for example write x[i:i+3,0] instead of x[i:i+3][0]")
    # TODO: this ckecking don't pass on Waterbucket.py, but the xml file is the same that the java version !
    # if options.checker:
    #    if not hasattr(Extension, "checked_tables"):
    #        Extension.checked_tables = set()
    #    if id(table) not in checked_tables:
    #        for t in table:
    #            for i, v in enumerate(t):
    #                if v not in variables[i].dom:
    #                    raise ValueError(
    #                        "Problem: Constraint extension : a value of table is not represented in a domain of a variable : " + str(domainElement))
    #        checked_tables.add(id(table))
    return ECtr(ConstraintExtension(scope, table, positive))


def Intension(node):
    checkType(node, Node)
    ctr = ECtr(ConstraintIntension(node))
    if ctr.blank_basic_attributes():
        ctr.copy_basic_attributes_of(node)
    node.mark_as_used()
    return ctr


def abs(*args):
    if len(args) == 1 and isinstance(args[0], Node) and args[0].type == TypeNode.SUB:
        return Node.build(TypeNode.DIST, *args[0].sons)
    return Node.build(TypeNode.ABS, *args) if len(args) == 1 and isinstance(args[0], (Node, Variable)) else absPython(*args)


def expr(operator, *args):
    return Node.build(operator, *args)


def min(*args):
    return Node.build(TypeNode.MIN, *args) if len(args) > 0 and any(isinstance(a, (Node, Variable)) for a in args) else minPython(*args)


def max(*args):
    return Node.build(TypeNode.MAX, *args) if len(args) > 0 and any(isinstance(a, (Node, Variable)) for a in args) else maxPython(*args)


def dist(*args):
    return Node.build(TypeNode.DIST, *args)


def xor(*args):
    return Node.build(TypeNode.XOR, *args) if len(args) > 1 else args[0]


def iff(*args):
    return Node.build(TypeNode.IFF, *args)


def imply(*args):
    return Node.build(TypeNode.IMP, *args)


def ift(*args):
    return Node.build(TypeNode.IF, *args)


def conjunction(*args):
    return Node.conjunction(*args)


def disjunction(*args):
    return Node.disjunction(*args)


def knight_attack(x, y, order):
    d1, d2 = dist(x // order, y // order), dist(x % order, y % order)  # distances between rows and columns
    return (d1 == 1) & (d2 == 2) | (d1 == 2) & (d2 == 1)


''' Language-based Constraints '''


def Regular(*, scope, automaton):
    scope = flatten(scope)
    checkType(scope, [Variable])
    checkType(automaton, Automaton)
    return ECtr(ConstraintRegular(scope, automaton))


def Mdd(*, scope, mdd):
    scope = flatten(scope)
    checkType(scope, [Variable])
    checkType(mdd, MDD)
    return ECtr(ConstraintMdd(scope, mdd.transitions))


''' Comparison-based Constraints '''


def AllDifferent(term, *others, excepting=None, matrix=None):
    terms = flatten(term, others)
    if len(terms) == 0 or (len(terms) == 1 and isinstance(terms[0], (int, Variable, Node))):
        return None
    if matrix is not None:
        assert excepting is None, "excepting values are currently not supported for AllDifferentMatrix"
        matrix = [flatten(row) for row in terms]
        assert all(len(row) == len(matrix[0]) for row in matrix), "The matrix id badly formed"
        assert all(checkType(l, [Variable]) for l in matrix)
        return ECtr(ConstraintAllDifferentMatrix(matrix))
    excepting = list(excepting) if isinstance(excepting, (tuple, set)) else [excepting] if isinstance(excepting, int) else excepting
    checkType(terms, ([Variable, Node]))
    checkType(excepting, ([int], type(None)))
    return ECtr(ConstraintAllDifferent(terms, excepting))


def AllDifferentList(term, *others, excepting=None):
    if isinstance(term, types.GeneratorType):
        term = [l for l in term]
    elif len(others) > 0:
        term = list((term,) + others)
    lists = [flatten(l) for l in term]
    assert all(checkType(l, [Variable]) for l in lists)
    excepting = list(excepting) if isinstance(excepting, (tuple, set)) else excepting
    checkType(excepting, ([int], type(None)))
    assert all(len(l) == len(lists[0]) for l in lists) and (excepting is None or len(excepting) == len(list[0]))
    return ECtr(ConstraintAllDifferentList(lists, excepting))


def AllEqual(term, *others):
    terms = flatten(term, others)
    checkType(terms, ([Variable], [Node]))
    return ECtr(ConstraintAllEqual(terms))


def _ordered(term, others, operator, lengths):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    checkType(operator, TypeOrderedOperator)
    checkType(lengths, ([int, Variable], type(None)))
    if lengths is not None:
        if len(terms) == len(lengths):
            lengths = lengths[:-1]  # we assume that the last value is useless
        assert len(terms) == len(lengths) + 1
    return ECtr(ConstraintOrdered(terms, operator, lengths))


def Increasing(term, *others, strict=False, lengths=None):
    return _ordered(term, others, TypeOrderedOperator.INCREASING if not strict else TypeOrderedOperator.STRICTLY_INCREASING, lengths)


def Decreasing(term, *others, strict=False, lengths=None):
    return _ordered(term, others, TypeOrderedOperator.DECREASING if not strict else TypeOrderedOperator.STRICTLY_DECREASING, lengths)


def _lex(term, others, operator, matrix):
    if len(others) == 0:
        assert is_matrix(term, Variable)
        lists = [flatten(l) for l in term]
    else:
        assert is_1d_list(term, Variable) and all(is_1d_list(l, Variable) for l in others)
        lists = [flatten(term)] + [flatten(l) for l in others]
    assert is_matrix(lists, Variable)  # new check because some null cells (variables) may have been discarded
    assert all(len(l) == len(lists[0]) for l in lists)
    checkType(lists, [Variable])
    checkType(operator, TypeOrderedOperator)
    return ECtr(ConstraintLexMatrix(lists, operator)) if matrix else ECtr(ConstraintLex(lists, operator))


def LexIncreasing(term, *others, strict=False, matrix=False):
    return _lex(term, others, TypeOrderedOperator.INCREASING if not strict else TypeOrderedOperator.STRICTLY_INCREASING, matrix)


def LexDecreasing(term, *others, strict=False, matrix=False):
    return _lex(term, others, TypeOrderedOperator.DECREASING if not strict else TypeOrderedOperator.STRICTLY_DECREASING, matrix)


''' Method for handling complete/partial constraints '''


def _wrapping_by_complete_or_partial_constraint(ctr):
    condition = ctr.arguments[TypeCtrArg.CONDITION].content
    return ECtr(ctr) if condition is not None else PartialConstraint(ctr)


''' Counting and Summing Constraints '''


def Sum(term, *others, condition=None):
    def _get_terms_coeffs(terms):
        if len(terms) == 1 and isinstance(terms[0], ScalarProduct):
            return flatten(terms[0].variables), flatten(terms[0].coeffs)
        if all(isinstance(x, (Variable, PartialConstraint)) for x in terms):
            return terms, None
        t1, t2 = [], []
        for tree in terms:
            if isinstance(tree, Variable):
                t1.append(tree)
                t2.append(1)
            else:
                assert isinstance(tree, (Node, NegVariable))
                pair = (tree.variable, -1) if isinstance(tree, NegVariable) else tree.tree_val_if_binary_type(TypeNode.MUL)
                if pair is None:
                    break
                t1.append(pair[0])
                t2.append(pair[1])
        if len(t1) == len(terms):
            for tree in terms:
                if isinstance(tree, Node):
                    tree.mark_as_used()
            return t1, t2
        return terms, None

    def _manage_coeffs(terms, coeffs):
        if coeffs:
            OpOverrider.disable()
            if len(coeffs) == 1 and isinstance(coeffs[0], (tuple, set, range)):
                coeffs = list(coeffs[0])
            elif isinstance(coeffs, (tuple, set, range)):
                coeffs = list(coeffs)
            elif isinstance(coeffs, (int, Variable)):
                coeffs = [coeffs]
            assert len(terms) == len(coeffs), "Lists (vars and coeffs) should have the same length. Here, we have " + str(len(terms)) + "!=" + str(len(coeffs))
            # if 0 in coeffs:
            #    terms = [term for i, term in enumerate(terms) if coeffs[i] != 0]
            #    coeffs = [coeff for coeff in coeffs if coeff != 0]
            # if all(c == 1 for c in coeffs): coeffs = None
            checkType(coeffs, ([Variable, int], type(None)))
            OpOverrider.enable()
        return terms, coeffs

    terms = list(term) if isinstance(term, types.GeneratorType) else flatten(term, others)
    checkType(terms, ([Variable], [Node], [PartialConstraint], [ScalarProduct]))
    auxiliary().replace_partial_constraints(terms)

    terms, coeffs = _get_terms_coeffs(terms)
    terms, coeffs = _manage_coeffs(terms, coeffs)
    # TODO control here some assumptions
    return _wrapping_by_complete_or_partial_constraint(ConstraintSum(terms, coeffs, Condition.build_condition(condition)))


def Count(term, *others, value=None, values=None, condition=None):
    terms = flatten(term, others)
    assert len(terms) > 0, "A count with an empty scope"
    checkType(terms, ([Variable], [Node]))
    if value is None and values is None:
        value = 1
    assert value is None or values is None, str(value) + " " + str(values)
    values = list(values) if isinstance(values, (tuple, set)) else [value] if isinstance(value, (int, Variable)) else values
    checkType(values, ([int], [Variable]))
    return _wrapping_by_complete_or_partial_constraint(ConstraintCount(terms, values, Condition.build_condition(condition)))


# def AtLeastOne(term, *others, value=None, values=None):
#     return Count(term, others, value=value, values=values) >= 1
#
# def AtMostOne(term, *others, value=None, values=None):
#     return Count(term, others, value=value, values=values) <= 1
#
# def ExactlyOne(term, *others, value=None, values=None):
#     return Count(term, others, value=value, values=values) == 1


def NValues(term, *others, excepting=None, condition=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    if excepting is not None:
        excepting = flatten(excepting)
        checkType(excepting, [int])
    return _wrapping_by_complete_or_partial_constraint(ConstraintNValues(terms, excepting, Condition.build_condition(condition)))


# def NotAllEqual(term, *others):
#    return NValues(term, others) > 1


def Cardinality(term, *others, occurrences, closed=False):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    assert isinstance(occurrences, dict)
    values = list(occurrences.keys())
    assert all(isinstance(value, (int, Variable)) for value in values)
    occurs = list(occurrences.values())
    checkType(closed, (bool, type(None)))
    for i, occ in enumerate(occurs):
        if isinstance(occ, range):
            occurs[i] = str(min(occ)) + ".." + str(max(occ))
        if isinstance(occ, list):
            flat = flatten(occ)
            if all([isinstance(e, int) for e in flat]) and flat == list(range(min(flat), max(flat) + 1)):
                flat = str(min(flat)) + ".." + str(max(flat))
            occurs[i] = flat
    return ECtr(ConstraintCardinality(terms, values, occurs, closed))


''' Connection Constraints '''


def _extremum(term, others, index, start_index, type_rank, condition, maximum):
    terms = list(term) if isinstance(term, types.GeneratorType) else flatten(term, others)
    terms = [Sum(t) if isinstance(t, ScalarProduct) else t for t in terms]  # to have PartialConstraints
    checkType(terms, ([Variable, Node], [PartialConstraint]))
    auxiliary().replace_partial_constraints(terms)
    checkType(index, (Variable, type(None)))
    checkType(start_index, int)
    checkType(type_rank, TypeRank)
    assert index is not None or (start_index == 0 and type_rank is TypeRank.ANY)
    return ConstraintMaximum(terms, index, start_index, type_rank, condition) if maximum else ConstraintMinimum(terms, index, start_index, type_rank, condition)


def Maximum(term, *others, index=None, start_index=0, type_rank=TypeRank.ANY, condition=None):
    return _wrapping_by_complete_or_partial_constraint(_extremum(term, others, index, start_index, type_rank, condition, True))


def Minimum(term, *others, index=None, start_index=0, type_rank=TypeRank.ANY, condition=None):
    return _wrapping_by_complete_or_partial_constraint(_extremum(term, others, index, start_index, type_rank, condition, False))


# def Element(*, vector, index=None, value, rank=TypeRank.ANY):
#    vector = flatten(vector)
#    checkType(vector, ([Variable], [int]))
#    checkType(index, (Variable, type(None)))
#    checkType(value, (Variable, int))
#    checkType(rank, TypeRank)
#    assert index is not None or rank is None, "rank is defined while index is not specified"
#    return ECtr(ConstraintElement(vector, value, index, rank if rank != TypeRank.ANY else None))


def Channel(list1, list2=None, *, start_index1=0, start_index2=0):
    list1 = flatten(list1)
    checkType(list1, [Variable])
    if list2:
        list2 = flatten(list2)
        checkType(list2, [Variable])
    checkType(start_index1, int)
    checkType(start_index2, int)
    assert start_index2 == 0 or list2 is not None, "start_index2 is defined while list2 is not specified"
    return ECtr(ConstraintChannel(list1, start_index1, list2, start_index2))


''' Packing and Scheduling Constraints '''


def NoOverlap(*, origins, lengths, zero_ignored=False):
    checkType(origins, [Variable])
    checkType(lengths, ([Variable], [int]))
    return ECtr(ConstraintNoOverlap(origins, lengths, zero_ignored))


def Cumulative(*, origins, lengths, ends=None, heights, condition=None):
    origins = flatten(origins)
    checkType(origins, [Variable])
    lengths = flatten(lengths)
    checkType(lengths, ([Variable], [int]))
    heights = flatten(heights)
    checkType(heights, ([Variable], [int]))
    if ends is not None: ends = flatten(ends)
    checkType(ends, ([Variable], type(None)))
    return _wrapping_by_complete_or_partial_constraint(ConstraintCumulative(origins, lengths, ends, heights, Condition.build_condition(condition)))


''' Constraints on Graphs'''


def Circuit(term, *others, start_index=0, size=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    checkType(start_index, int)
    checkType(size, (int, type(None)))
    return ECtr(ConstraintCircuit(terms, start_index, size))


''' Elementary Constraints '''


def Clause(term, *others, phases=None):
    literals = flatten(term, others)
    phases = [False] * len(literals) if phases is None else flatten(phases)
    assert len(literals) == len(phases)
    checkType(literals, ([Variable, NotVariable]))
    checkType(phases, [bool])
    for i, literal in enumerate(literals):
        if isinstance(literal, NotVariable):
            literals[i] = literal.variable
            phases[i] = True
    return ECtr(ConstraintClause(literals, phases))


''' Objectives '''


def _optimize(term, minimization):
    if isinstance(term, ScalarProduct):
        term = Sum(term)  # to have a PartialConstraint
    checkType(term, (Variable, Node, PartialConstraint)), "Did you forget to use Sum, e.g., as in Sum(x[i]*3 for i in range(10))"
    satisfy(pc == var for (pc, var) in auxiliary().collected())
    comment, _, tag, _ = comments_and_tags_of_parameters_of(function_name="minimize" if minimization else "maximize", args=[term])
    type = TypeCtr.MINIMIZE if minimization else TypeCtr.MAXIMIZE
    if isinstance(term, (Variable, Node)):
        if isinstance(term, Node):
            term.mark_as_used()
        return EObjective(ObjectiveExpression(type, term)).note(comment[0]).tag(tag[0])  # TODO why only one tag?
    else:
        return EObjective(ObjectivePartial(type, term)).note(comment[0]).tag(tag[0])


def minimize(term):
    return _optimize(term, True)


def maximize(term):
    return _optimize(term, False)


''' Annotations '''


def annotate(*, decision=None, output=None, varHeuristic=None, valHeuristic=None, filtering=None, prepro=None, search=None, restarts=None):
    def add_annotation(obj, Ann):
        if obj:
            ann = Ann(obj)
            assert type(ann) not in AnnEntities.items_types, "This type of annotation can be specified only one time"
            annotations.append(EAnnotation(ann))

    annotations = []
    add_annotation(decision, AnnotationDecision)
    add_annotation(output, AnnotationOutput)
    add_annotation(varHeuristic, AnnotationVarHeuristic)
    add_annotation(valHeuristic, AnnotationValHeuristic)
    add_annotation(filtering, AnnotationFiltering)
    add_annotation(prepro, AnnotationPrepro)
    add_annotation(search, AnnotationSearch)
    add_annotation(restarts, AnnotationRestarts)
    return annotations


''' Helpers '''


# def column(m, j):
#     assert is_2d_list(m), "column() can only be called on 2-dimensional lists"
#     assert all(len(row) > j for row in m), "one row has not at least j+1 elements"
#     return ListVar(row[j] for row in m)
#
#
# def columns(m):
#     assert is_matrix(m), "columns() can only be called on matrices"
#     return ListVar(column(m, j) for j in range(len(m[0])))


def diagonal_down(m, i=-1, j=-1, check=True):
    if check is True:
        assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if i == -1 and j == -1:
        return diagonal_down(m, 0, 0, False)
    if j == -1:
        return ListVar(m[k][len(m) - (i - k) if k < i else k - i] for k in range(len(m)))
    return ListVar(m[i + k][j + k] for k in range(len(m) - max(i, j)))


def diagonals_down(m, *, broken=False):
    assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if broken:
        return ListVar(diagonal_down(m, i, -1, False) for i in range(len(m)))
    return ListVar(diagonal_down(m, i, 0, False) for i in reversed(range(len(m) - 1))) + ListVar(diagonal_down(m, 0, j, False) for j in range(1, len(m) - 1))


def diagonal_up(m, i=-1, j=-1, check=True):
    if check is True:
        assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if i == -1 and j == -1:
        return diagonal_up(m, len(m) - 1, 0, False)
    if j == -1:
        return ListVar(m[k][len(m) - i - k - 1 if k < len(m) - i else 2 * len(m) - i - k - 1] for k in range(len(m)))
    return ListVar(m[i - k][j + k] for k in range(min(i + 1, len(m) - j)))


def diagonals_up(m, *, broken=False):
    assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if broken:
        return ListVar(diagonal_up(m, i, -1, False) for i in range(len(m)))
    return ListVar(diagonal_up(m, i, 0, False) for i in range(1, len(m))) + ListVar(diagonal_up(m, len(m) - 1, j, False) for j in range(1, len(m) - 1))


def different_values(*args):
    assert all(isinstance(arg, int) for arg in args)
    return all(a != b for (a, b) in combinations(args, 2))


def to_ordinary_table(table, domains, *, keep_any=False):
    doms = [range(d) if isinstance(d, int) else d.all_values() if isinstance(d, Domain) else d for d in domains]
    tbl = set()
    if keep_any:
        for t in table:
            if any(isinstance(v, ConditionValue) for v in t):  # v may be a ConditionValue (with method 'filtering')
                tbl.update(product(*({v} if isinstance(v, int) or v == ANY else v.filtering(doms[i]) for i, v in enumerate(t))))
            else:
                tbl.add(t)
    else:
        for t in table:
            if any(v == ANY or isinstance(v, ConditionValue) for v in t):  # v may be a ConditionValue (with method 'filtering')
                tbl.update(product(*({v} if isinstance(v, int) else doms[i] if v == ANY else v.filtering(doms[i]) for i, v in enumerate(t))))
            else:
                tbl.add(t)
    return tbl


def cp_array(*l):
    if len(l) == 1:
        l = l[0]
    if isinstance(l, (tuple, set, types.GeneratorType)):
        l = list(l)
    assert isinstance(l, list) and len(l) > 0
    if isinstance(l[0], (list, types.GeneratorType)):
        assert all(isinstance(t, (list, types.GeneratorType)) for t in l)
        res = [cp_array(t) for t in l]
        return ListInt(res) if isinstance(res[0], ListInt) else ListVar(res)
    if all(isinstance(v, int) for v in l):
        return ListInt(l)
    elif all(isinstance(v, Variable) for v in l):
        return ListVar(l)
    else:
        raise NotImplemented


def _pycharm_security():  # for avoiding that imports are removed when reformatting code
    _ = (permutations, transpose, alphabet_positions, all_primes, integer_scaling, columns, namedtuple, default_data, ne)



"""
See https://en.wikipedia.org/wiki/Flow_shop_scheduling

Example of Execution:
  python3 SchedulingFS.py -data=SchedulingFS-Taillard-os-04-04-0.json
"""

from pycsp3 import *

durations = data  # durations[i][j] is the duration of operation/machine j for job i
horizon = sum(sum(t) for t in durations) + 1
n, m = len(durations), len(durations[0])

# s[i][j] is the start time of the jth operation for the ith job
s = VarArray(size=[n, m], dom=range(horizon))

satisfy(
    # operations must be ordered on each job
    [Increasing(s[i], lengths=durations[i]) for i in range(n)],

    # no overlap on resources
    [NoOverlap(origins=s[:, j], lengths=durations[:, j]) for j in range(m)]
)

minimize(
    # minimizing the makespan
    Maximum(s[i][- 1] + durations[i][- 1] for i in range(n))
)



"""
Five friends (three women named Mindy, Sabrina and Tanya and two men named Anton and Chet) agreed to meet on an island
in the middle of the bay for a picnic lunch on Sunday.
Each arrived on a different type of conveyance (jet-ski, kayak, outrigger canoe, sailboard, and sailboat),
and each brought a different beverage (beer, juice, soda, water and wine)
and a different food (bread, cheese, cookies, fish and salad) to share with the others.
We know that:
 - Mindy and the person who arrived on the jet-ski brought, in some order, beer and soda.
 - The five people are Chet, the person who arrived by kayak, and the people who brought the fish, water and wine.
 - The person who brought the salad didn't bring the wine.
 - Sabrina and the person who brought the cheese are, in some order, the person who arrived by outrigger canoe and the one who brought the beer.
 - A woman brought bread and juice.
 - The person who arrived by sailboat is not the person who brought the soda and the other person who brought the cookies.

Execution:
  python3 Picnic.py
"""

from pycsp3 import *

friends = Mindy, Sabrina, Tanya, Anton, Chet = "Mindy", "Sabrina", "Tanya", "Anton", "Chet"

# conveyances[i] is the friend who arrived with the ith conveyance
jetski, kayak, canoe, sailboard, sailboat = conveyances = VarArray(size=5, dom=friends)

# drinks[i] is the friend who brought the ith drink
beer, juice, soda, water, wine = drinks = VarArray(size=5, dom=friends)

# foods[i] is the friend who brought the ith food
bread, cheese, cookies, fish, salad = foods = VarArray(size=5, dom=friends)

satisfy(
    AllDifferent(conveyances),
    AllDifferent(drinks),
    AllDifferent(foods),

    (beer == Mindy) | (soda == Mindy),
    (beer == jetski) | (soda == jetski),
    jetski != Mindy,
    kayak != Chet,
    fish != Chet,
    water != Chet,
    wine != Chet,
    AllDifferent(kayak, fish, water, wine),
    salad != wine,
    (canoe == Sabrina) | (beer == Sabrina),
    (canoe == cheese) | (beer == cheese),
    canoe != beer,
    cheese != Sabrina,
    jetski in {beer, soda},
    Mindy in {beer, soda},
    bread in {Mindy, Sabrina, Tanya},
    juice in {Mindy, Sabrina, Tanya},
    bread == juice,
    sailboat not in {soda, cookies},
    soda != cookies
)



"""
See "Search strategies for rectangle packing" by H. Simonis and B. O'Sullivan, CP 2008.
See CP papers on short supports.

Example of Execution:
  python3 RectPacking.py -data=RectPacking_perfect-001.json
"""

import math

from pycsp3 import *

width, height = data.container
boxes = data.boxes
nBoxes = len(boxes)

# x[i] is the x-coordinate where is put the ith rectangle
x = VarArray(size=nBoxes, dom=range(width))

# y[i] is the y-coordinate where is put the ith rectangle
y = VarArray(size=nBoxes, dom=range(height))

satisfy(
    # unary constraints on x
    [x[i] + boxes[i].width <= width for i in range(nBoxes)],

    # unary constraints on y
    [y[i] + boxes[i].height <= height for i in range(nBoxes)],

    # no overlap on boxes
    NoOverlap(origins=[(x[i], y[i]) for i in range(nBoxes)], lengths=[(w, h) for (w, h) in boxes]),

    # tag(symmetry-breaking)
    [x[- 1] <= math.floor((width - boxes[- 1].width) // 2.0), y[- 1] <= x[- 1]] if width == height else None
)



from pycsp3.problems.tests.tester import Tester, run

NAME = "g6_testing"

xcsp = (Tester(NAME)
         # .add("UnitTestingIntension")
         .add("TestAbscon")
         .add("TestElement")
         .add("TestSlices")
         .add("TestSumConditions")
         .add("TestSums")
         .add("TestUnaryConditions")
         .add("UnitTestingExtension")
         .add("UnitTestingSum")
         .add("UnitTestingVariable")
         )

run(xcsp)



"""
The Dreadsbury Mansion Mystery (4 solutions)
F. J. Pelletier, "Seventy-five problems for testing automatic theorem provers",
Journal of Automated Reasoning, 2: 191 216, 1986.

Someone who lives in Dreadsbury Mansion killed Aunt Agatha.
Agatha, the butler, and Charles live in Dreadsbury Mansion, and are the only people who live therein.
 - A killer always hates his victim, and is never richer than his victim.
 - Charles hates no one that Aunt Agatha hates.
 - Agatha hates everyone except the butler.
 - The butler hates everyone not richer than Aunt Agatha.
 - The butler hates everyone Agatha hates.
 - No one hates everyone.

Execution:
  python3 Agatha.py
"""

from pycsp3 import *

persons = agatha, butler, charles = 0, 1, 2

# killer is the person who kills Agatha
killer = Var(dom=persons)

# hating[i][j] is 1 iff person i hates person j
hating = VarArray(size=[3, 3], dom={0, 1})

# richer[i][j] is 1 iff person i is richer than person j
richer = VarArray(size=[3, 3], dom={0, 1})

satisfy(
    # a killer always hates his victim
    hating[killer][agatha] == 1,

    # a killer is never richer than his victim
    richer[killer][agatha] == 0,

    # Charles hates no one that Agatha hates
    [imply(hating[agatha][p], ~hating[charles][p]) for p in persons],

    # Agatha hates everybody except the butler
    [hating[agatha][p] == 1 for p in persons if p != butler],

    # the butler hates everyone not richer than Aunt Agatha
    [imply(~richer[p, agatha], hating[butler, p]) for p in persons],

    # the butler hates everyone Agatha hates
    [imply(hating[agatha, p], hating[butler, p]) for p in persons],

    # no one hates everyone
    [Count(hating[p], value=0) > 0 for p in persons]
)



"""
See https://en.wikipedia.org/wiki/Knapsack_problem

Example of Execution:
  python3 Knapsack.py -data=Knapsack_20-50-00.json
"""

from pycsp3 import *

capacity, items = data
weights, values = zip(*items)
nItems = len(items)

# x[i] is 1 iff the ith item is selected
x = VarArray(size=nItems, dom={0, 1})

satisfy(
    # not exceeding the capacity of the knapsack
    x * weights <= capacity
)

maximize(
    # maximizing summed up value (benefit)
    x * values
)



"""
From the Minizinc tutorial

Execution:
  python3 Recipe.py
"""

from pycsp3 import *

# b is the number of banana cakes
b = Var(range(100))

# c is the number of chocolate cakes
c = Var(range(100))

satisfy(
    # using at most 4000 grams of flour
    250 * b + 200 * c <= 4000,

    # using at most 6 bananas
    2 * b <= 6,

    # using at most 2000 grams of sugar
    75 * b + 150 * c <= 2000,

    # using at most 500 grams of butter
    100 * b + 150 * c <= 500,

    # using at most 500 grams of cocoa
    75 * c <= 500
)

maximize(
    # maximizing the profit (400 and 450 cents for each banana and chocolate cake, respectively)
    b * 400 + c * 450
)



"""
See https://en.wikipedia.org/wiki/Change-making_problem

Examples of Execution:
  python3 ChangeMaking.py
  python3 ChangeMaking.py -data=10
  python3 ChangeMaking.py -data=10 -variant=compact
"""

from pycsp3 import *

k = data or 13

if not variant():
    # c1 is the number of coins of 1 cent
    c1 = Var(range(50))

    # c5 is the number of coins of 5 cents
    c5 = Var(range(50))

    # c10 is the number of coins of 10 cents
    c10 = Var(range(50))

    # c20 is the number of coins of 20 cents
    c20 = Var(range(50))

    # c50 is the number of coins of 50 cents
    c50 = Var(range(50))

    # e1 is the number of coins of 1 euro
    e1 = Var(range(50))

    # e2 is the number of coins of 2 euros
    e2 = Var(range(50))

    satisfy(
        # the given change must be correct
        [c1, c5, c10, c20, c50, e1, e2] * [1, 5, 10, 20, 50, 100, 200] == k
    )

    minimize(
        # the given change must have the minimum number of coins
        c1 + c5 + c10 + c20 + c50 + e1 + e2
    )

elif variant("compact"):
    # coins[i] is the number of coins of the ith type
    c1, c5, c10, c20, c50, e1, e2 = coins = VarArray(size=7, dom=range(50))

    satisfy(
        # the given change must be correct
        coins * [1, 5, 10, 20, 50, 100, 200] == k
    )

    minimize(
        # the given change must have the minimum number of coins
        Sum(coins)
    )



import types
# import itertools
import numpy
import re

from enum import Enum, unique
from functools import reduce
from itertools import product

from pycsp3.classes.auxiliary.ptypes import auto
from pycsp3.classes.main.variables import Variable, NotVariable, NegVariable
from pycsp3.classes import main
from pycsp3.dashboard import options
from pycsp3.tools.utilities import flatten, is_containing
from pycsp3.tools.inspector import checkType


class Entity:
    def __init__(self, name, comment=None, tags=[]):
        self.id = name
        self.comment = None
        self.note(comment)  # we use comment instead of note because we need method note()
        self.tags = []
        self.tag(tags)

    def note(self, comment):
        if comment is not None and comment.strip() != "":
            self.comment = comment.strip() if self.comment is None else self.comment + " - " + comment.strip()
        return self

    def tag(self, tags):
        if tags is not None:
            toks = (tok.strip() for tok in tags.strip().split(" ")) if isinstance(tags, str) else (tok.strip() for tok in tags)
            self.tags.extend([tok for tok in toks if tok != "" and tok not in self.tags])
        return self

    def same_type_and_basic_attributes(self, other):
        return type(self) == type(other) and self.comment == other.comment and self.tags == other.tags

    def mergeable_with(self, other):
        return (type(self) == type(other) and (self.comment == other.comment or None in {self.comment, other.comment})
                and (self.tags == other.tags or 0 in {len(self.tags), len(other.tags)}))

    def blank_basic_attributes(self):
        return self.comment is None and self.tags == []

    def clear_basic_attributes(self):
        self.comment = None
        self.tags = []

    def copy_basic_attributes_of(self, other):
        assert isinstance(other, Entity)
        self.comment = other.comment
        self.tags = other.tags
        return self


class EVar(Entity):
    def __init__(self, x, comment=None, tags=[]):
        super().__init__(x.id, comment, tags)
        self.variable = x
        VarEntities.items.append(self)
        VarEntities.varToEVar[x] = self

    def get_type(self):
        return self.variable.dom.get_type()

    def __call__(self):
        return self.variable


class EVarArray(Entity):
    def __init__(self, X, name, comment=None, tags=[]):
        super().__init__(name, comment, tags)
        self.name = name
        self.variables = X
        self.flatVars = flatten(X)
        assert len(self.flatVars) != 0, "Array of variable empty !"
        self.containing_hole = None  # undefined until we ask  #flatVarsKeepingNone = flatten(X, keep_none=True)
        self.size = []
        curr = self.variables
        while isinstance(curr, list):
            self.size.append(len(curr))
            curr = curr[0]
        VarEntities.items.append(self)
        for x in self.flatVars:
            VarEntities.varToEVarArray[x] = self
        VarEntities.prefixToEVarArray[name] = self

    def is_containing_hole(self):
        if self.containing_hole is None:
            self.containing_hole = is_containing(self.variables, type(None), check_first_only=True)
        return self.containing_hole

    def extend_with(self, var):  # used when building auxiliary variables (to be used with global constraints)
        self.variables.append(var)
        self.flatVars.append(var)
        self.size[0] += 1
        VarEntities.varToEVarArray[var] = self

    def get_type(self):
        return self.flatVars[0].dom.get_type()

    def __getitem__(self, i):
        return self.variables[i]

    def __len__(self):
        return len(self.variables)

    def __iter__(self):
        yield self.variables.__iter__()

    def __next__(self):
        return self.variables.__next__()

    def size_to_string(self):
        return "".join("[" + str(v) + "]" for v in self.size)


""" Class to represent stand-alone constraints """


class ECtr(Entity):
    def __init__(self, c):
        super().__init__(None)  # no need to have an id here
        if c is None:
            self.constraint = None
            print("Warning: a constraint is None")
        else:
            self.constraint = c
            # CtrEntities.allEntities.append(self)


class ECtrs(Entity):
    """ Class for representing sets of constraints """

    def __init__(self, constraints=None):
        super().__init__(None)  # no need to have an id here
        assert constraints is not None and isinstance(constraints, list)
        self.entities = [c for c in constraints if c is not None]


class EToGather(ECtrs):
    ''' Constraints possibly stored in a group (the user asked to gather these constraints)'''

    def __init__(self, constraints):
        super().__init__(constraints)


class EToSatisfy(ECtrs):
    ''' Constraints possibly stored in several groups or several blocks (block built when a group is not possible) or stand-alone constraints'''

    def __init__(self, constraints):
        assert constraints is not None
        constraints = [c for c in constraints if c is not None]
        if len(constraints) > 0:
            CtrEntities.items.append(self)
            super().__init__(constraints)


class EGroup(ECtrs):
    ''' Constraints in a group '''

    def __init__(self):
        super().__init__([])
        self.abstraction = ""
        self.all_args = []


class EBlock(ECtrs):
    def __init__(self, constraints):
        super().__init__(constraints)


class ESlide(ECtrs):
    ''' Constraints possibly stored as a slide meta-constraint (the user asked to slide the constraints)'''

    def __init__(self, constraints):
        super().__init__(constraints)
        self.scope = []
        self.offset = False
        self.circular = False


class EIfThenElse(ECtrs):
    def __init__(self, constraints):
        checkType(constraints, [ECtr])
        assert len(constraints) == 3, "Error: three components must be specified in ifThenElse"
        super().__init__(constraints)


class EObjective(Entity):
    def __init__(self, c):
        if c is None:
            return
        super().__init__(None)  # no need to have an id here
        self.constraint = c
        ObjEntities.items.append(self)


class EAnnotation(Entity):
    def __init__(self, c):
        if c is None:
            return
        super().__init__(None)  # no need to have an id here
        self.constraint = c
        AnnEntities.items.append(self)
        AnnEntities.items_types.append(type(c))


class VarEntities:
    items = []
    varToEVar = dict()
    varToEVarArray = dict()
    prefixToEVarArray = dict()

    @staticmethod
    def get_item_with_name(s):
        if '[' in s:  # we need to look for arrays
            pos = s.index("[")
            prefix, suffix = s[:pos], s[pos:]
            assert prefix in VarEntities.prefixToEVarArray
            va = VarEntities.prefixToEVarArray[prefix]
            indexes = [int(v) if len(v) > 0 else None for v in re.split("\]\[", suffix[1:-1])]
            if is_containing(indexes, int):
                res = va.variables
                for v in indexes:
                    res = res[v]
                return res
            else:
                assert is_containing(indexes, type(None))
                return va
        else:
            for item in VarEntities.items:
                if isinstance(item, EVar) and item.id == s:
                    return item
        return None


class CtrEntities:
    items = []


class ObjEntities:
    items = []


class AnnEntities:
    items = []
    items_types = []


@unique
class TypeNode(Enum):
    def __init__(self, id, min_arity, max_arity):
        self.id = id
        self.min_arity = min_arity
        self.max_arity = max_arity
        self.lowercase_name = self.name.lower()

    def __str__(self):
        return self.lowercase_name

    ''' 0-ary '''
    VAR, PAR, INT, RATIONAL, DECIMAL, SYMBOL, PARTIAL = ((id, 0, 0) for id in auto(7))

    ''' Unary'''
    NEG, ABS, SQR, NOT, CARD, HULL, CONVEX, SQRT, EXP, LN, SIN, COS, TAN, ASIN, ACOS, ATAN, SINH, COSH, TANH = ((id, 1, 1) for id in auto(19))

    ''' Binary '''
    SUB, DIV, MOD, POW, DIST, LT, LE, GE, GT, IN, NOTIN, IMP, DIFF, DJOINT, SUBSET, SUBSEQ, SUPSEQ, SUPSET, FDIV, FMOD, = ((id, 2, 2) for id in auto(20))

    ''' Ternary '''
    IF = (auto(), 3, 3)

    ''' N-ary (2 to infinity)'''
    ADD, MUL, MIN, MAX, NE, EQ, AND, OR, XOR, IFF, UNION, INTER, SDIFF = ((id, 2, float("inf")) for id in auto(13))

    SET = (auto(), 0, float("inf"))

    SPECIAL = (auto(), 0, float("inf"))

    def is_leaf(self):
        return self == TypeNode.SPECIAL or (self.min_arity == self.max_arity == 0)

    def is_valid_arity(self, k):
        return self.min_arity <= k <= self.max_arity

    def is_logical_operator(self):
        return self in {TypeNode.NOT, TypeNode.AND, TypeNode.OR, TypeNode.XOR, TypeNode.IFF, TypeNode.IMP}

    def is_relational_operator(self):
        return self in {TypeNode.LT, TypeNode.LE, TypeNode.GE, TypeNode.GT, TypeNode.EQ, TypeNode.NE}

    def is_predicate_operator(self):
        return self.is_logical_operator() or self.is_relational_operator() or self in {TypeNode.IN, TypeNode.NOTIN}


def neg_range(r):
    assert isinstance(r, range) and r.step == 1
    return range(-r.stop + 1, -r.start + 1)


def abs_range(r):
    assert isinstance(r, range) and r.step == 1
    return range(0 if 0 in r else min(abs(r.start), abs(r.stop - 1)), max(abs(r.start), abs(r.stop - 1)) + 1)


def add_range(r1, r2):
    assert isinstance(r1, range) and r1.step == 1 and isinstance(r2, range) and r2.step == 1
    return range(r1.start + r2.start, r1.stop + r2.stop - 1)


def possible_range(s, control_int=False):
    assert isinstance(s, set) and (not control_int or all(isinstance(v, int) for v in s))
    l = sorted(s)
    return range(l[0], l[-1] + 1) if 1 < l[-1] - l[0] + 1 == len(l) else l


class Node(Entity):
    all_nodes = []

    def __init__(self, type, args):
        super().__init__(None)
        Node.all_nodes.append(self)
        self.used = False
        self.type = type
        self.leaf = type.is_leaf()
        self.sons = args  # TODO sons is used whatever this is a parent or a leaf node; not a good choice. change the name of this field ??? to content ??
        self.abstractTree = None
        self.abstractValues = None

    def __str__(self):
        return str(self.sons) if self.type.is_leaf() else str(self.type) + "(" + ",".join(str(son) for son in self.sons) + ")"

    def possible_values(self):
        if self.type.is_predicate_operator():
            return range(0, 2)  # we use a range instead of [0,1] because it simplifies computation (see code below)
        if self.type.min_arity == self.type.max_arity == 0:
            if self.type == TypeNode.VAR:
                av = self.sons.dom.all_values()  # either a range or a sorted list of integers is returned
                if isinstance(av, range):
                    return av
                return range(av[0], av[0] + 1) if len(av) == 1 else range(av[0], av[1] + 1) if len(av) == 2 and av[0] + 1 == av[1] else av
            if self.type == TypeNode.INT:
                return range(self.sons, self.sons+1)  # we use a range instead of a singleton list because it simplifies computation (see code below)
            assert False, "no such 0-ary type " + str(self.type) + " is expected"
        if self.type.min_arity == self.type.max_arity == 1:
            pv = self.sons[0].possible_values()
            if self.type == TypeNode.NEG:
                return neg_range(pv) if isinstance(pv, range) else [-v for v in reversed(pv)]
            if self.type == TypeNode.ABS:
                return abs_range(pv) if isinstance(pv, range) else possible_range({abs(v) for v in pv})
            if self.type == TypeNode.SQR:
                return possible_range({v * v for v in pv})
            assert False, "no such 1-ary type " + str(self.type) + " is expected"
        if self.type.min_arity == self.type.max_arity == 2:
            pv1, pv2 = self.sons[0].possible_values(), self.sons[1].possible_values()
            all_ranges = isinstance(pv1, range) and isinstance(pv2, range)
            if self.type == TypeNode.SUB:
                return add_range(pv1, neg_range(pv2)) if all_ranges else possible_range({v1 - v2 for v1 in pv1 for v2 in pv2})
            if self.type == TypeNode.DIV:
                return possible_range({v1 // v2 for v1 in pv1 for v2 in pv2})
            if self.type == TypeNode.MOD:
                return possible_range({v1 % v2 for v1 in pv1 for v2 in pv2})
            if self.type == TypeNode.POW:
                return possible_range({v1 ** v2 for v1 in pv1 for v2 in pv2}, control_int=True)
            if self.type == TypeNode.DIST:
                return abs_range(add_range(pv1, neg_range(pv2))) if all_ranges else possible_range({abs(v1 - v2) for v1 in pv1 for v2 in pv2})
            assert False, "no such 2-ary type " + str(self.type) + " is expected"
        if self.type == TypeNode.IF:
            pv1, pv2 = self.sons[1].possible_values(), self.sons[2].possible_values()  # sons[0] is for the condition
            if isinstance(pv1, range) and isinstance(pv2, range) and len(range(max(pv1.start, pv2.start), min(pv1.stop, pv2.stop))) > 0:
                return range(min(pv1.start, pv2.start), max(pv1.stop, pv2.stop))
            return possible_range({v1 for v1 in pv1} | {v2 for v2 in pv2})
        if self.type.min_arity == 2 and self.type.max_arity == float("inf"):
            pvs = [son.possible_values() for son in self.sons]
            all_ranges = all(isinstance(pv, range) for pv in pvs)
            if self.type == TypeNode.ADD:
                return reduce(add_range, pvs) if all_ranges else possible_range({sum(p) for p in product(*(pv for pv in pvs))})
            if self.type == TypeNode.MUL:
                return possible_range({numpy.prod(p) for p in product(*(pv for pv in pvs))})
            # TODO: in case of all_ranges being False, possibility of improving the efficiency of the code below for MIN and MAX
            if self.type == TypeNode.MIN:
                return range(min(pv.start for pv in pvs), min(pv.stop for pv in pvs)) if all_ranges \
                    else possible_range({min(p) for p in product(*(pv for pv in pvs))})
            if self.type == TypeNode.MAX:
                return range(max(pv.start for pv in pvs), max(pv.stop for pv in pvs)) if all_ranges \
                    else possible_range({max(p) for p in product(*(pv for pv in pvs))})
        assert False, "The operator " + str(self.type) + " currently not implemented"

    def mark_as_used(self):
        self.used = True
        if isinstance(self.sons, list):
            for son in self.sons:
                Node.mark_as_used(son)

    def _abstraction_recursive(self, cache, harvest_values):
        if self.type in {TypeNode.VAR, TypeNode.INT, TypeNode.SYMBOL}:
            key = id(self)
            if key not in cache:
                cache[key] = len(harvest_values)  # can it be a problem to use it as a key?
                harvest_values.append(self.sons)
            return "%" + str(cache[key]), harvest_values
        return str(self.type) + "(" + ",".join(son._abstraction_recursive(cache, harvest_values)[0] for son in self.sons) + ")", harvest_values

    def _abstraction(self):
        if self.abstractTree is None:
            self.abstractTree, self.abstractValues = self._abstraction_recursive(dict(), [])

    def abstract_tree(self):
        self._abstraction()
        return self.abstractTree

    def abstract_values(self):
        self._abstraction()
        return self.abstractValues

    def _variables_recursive(self, harvest):
        if isinstance(self.sons, list):
            for son in self.sons:
                son._variables_recursive(harvest)
        if self.leaf and self.type == TypeNode.VAR:
            if self.sons not in harvest:
                harvest.add(self.sons)
        return harvest

    def variables(self):
        return self._variables_recursive([])

    def variable(self, i):
        return self.variables()[i]

    def flatten_by_associativity(self, node_type):
        while True:
            for i, son in enumerate(self.sons):
                if self.type == son.type == node_type:
                    self.sons.pop(i)
                    for s in reversed(son.sons):
                        self.sons.insert(i, s)
                    break
            else:  # no break
                break

    def reduce_integers(self):
        if self.type not in {TypeNode.ADD, TypeNode.MUL}:
            return
        ints, sons = [], []
        for son in self.sons:
            if son.type == TypeNode.INT:
                ints.append(son.sons)
            else:
                sons.append(son)
        if len(ints) > 1:
            value = reduce(lambda x, y: x + y, ints, 0) if self.type == TypeNode.ADD else reduce(lambda x, y: x * y, ints, 1)
            sons.append(Node(TypeNode.INT, value))
            self.sons = sons

    def var_val_if_binary_type(self, t):
        if self.type != t or len(self.sons) != 2 or self.sons[0].type == self.sons[1].type:
            return None
        if self.sons[0].type == TypeNode.VAR and self.sons[1].type == TypeNode.INT:
            return self.sons[0].sons, self.sons[1].sons
        elif self.sons[0].type == TypeNode.INT and self.sons[1].type == TypeNode.VAR:
            return self.sons[1].sons, self.sons[0].sons
        else:
            return None

    def tree_val_if_binary_type(self, t):
        if self.type != t or len(self.sons) != 2 or self.sons[0].type == self.sons[1].type:
            return None
        if self.sons[0].type != TypeNode.INT and self.sons[1].type == TypeNode.INT:
            return self.sons[0].sons if self.sons[0].type == TypeNode.VAR else self.sons[0], self.sons[1].sons
        elif self.sons[0].type == TypeNode.INT and self.sons[1].type != TypeNode.INT:
            return self.sons[1].sons if self.sons[1].type == TypeNode.VAR else self.sons[1], self.sons[0].sons
        else:
            return None

    """
      Static methods
    """

    @staticmethod
    def _create_sons(*args):
        t = []
        for arg in args:
            if isinstance(arg, Node):
                t.append(arg)
            elif isinstance(arg, EVar):
                t.append(Node(TypeNode.VAR, arg.variable))
            elif isinstance(arg, NotVariable):
                t.append(Node(TypeNode.NOT, [Node(TypeNode.VAR, arg.variable)]))
            elif isinstance(arg, NegVariable):
                t.append(Node(TypeNode.NEG, [Node(TypeNode.VAR, arg.variable)]))
            elif isinstance(arg, Variable):
                t.append(Node(TypeNode.VAR, arg))
            elif isinstance(arg, int):
                t.append(Node(TypeNode.INT, arg))
            elif isinstance(arg, str):
                t.append(Node(TypeNode.SYMBOL, arg))
            elif isinstance(arg, main.constraints.PartialConstraint):
                t.append(Node(TypeNode.PARTIAL, arg))
            else:
                raise ValueError("Problem: bad form of predicate " + str(arg))
        return t

    @staticmethod
    def build(type, *args):
        if isinstance(type, str):
            if type == "<":
                type = TypeNode.LT
            elif type == "<=":
                type = TypeNode.LE
            elif type == ">=":
                type = TypeNode.GE
            elif type == ">":
                type = TypeNode.GT
            elif type in {"=", "=="}:
                type = TypeNode.EQ
            elif type == "!=":
                type = TypeNode.NE
            else:
                type = TypeNode[type.upper()]
        if type is TypeNode.SET:
            assert len(args) == 1
            elements = list(args[0])
            sorted_sons = sorted(elements, key=lambda v: str(v)) if len(elements) > 0 and not isinstance(elements[0], int) else sorted(elements)
            return Node(type, Node._create_sons(*sorted_sons))  # *sorted(args[0])))
        args = flatten(Node.build(TypeNode.SET, arg) if isinstance(arg, (set, range, frozenset)) else arg for arg in args)
        assert type.is_valid_arity(len(args)), "Problem: Bad arity for node " + type.name + ". It is " + str(
            len(args)) + " but it should be between " + str(type.arityMin) + " and " + str(type.arityMax)
        node = Node(type, Node._create_sons(*args))
        if type == TypeNode.EQ and all(son.type.is_predicate_operator() for son in node.sons):
            node = Node(TypeNode.IFF, node.sons)
        # Reducing the node
        for t in {TypeNode.ADD, TypeNode.MUL, TypeNode.OR, TypeNode.AND, TypeNode.EQ, TypeNode.IFF}:
            node.flatten_by_associativity(t)
        node.reduce_integers()
        return node

    @staticmethod
    def set(*args):
        return Node.build(TypeNode.SET, *args)

    @staticmethod
    def _and_or(t, *args):
        assert t in {TypeNode.AND, TypeNode.OR}
        if len(args) == 1:
            if isinstance(args[0], list):
                args = tuple(args[0])
            if isinstance(args[0], types.GeneratorType):
                args = tuple(list(args[0]))
        return Node.build(t, *args) if len(args) > 1 else args[0]

    @staticmethod
    def conjunction(*args):
        return Node._and_or(TypeNode.AND, *args)

    @staticmethod
    def disjunction(*args):
        return Node._and_or(TypeNode.OR, *args)



"""
See "A Class of Hard Small 0-1 Programs" by Gérard Cornuéjols and Milind Dawande, INFORMS J. Comput. 11(2): 205-210 (1999)
See "

The feasibility problem consists of m 0-1 equality knapsack constraints defined on the same set of 10(m−1) variables.

Example of Execution:
  python3 MarketSplit.py -data=MarketSplit_04.json
"""

from pycsp3 import *

n, constraints = data

x = VarArray(size=n, dom={0, 1})

satisfy(
    x * coeffs == k for (coeffs, k) in constraints
)



"""
See https://en.wikipedia.org/wiki/Verbal_arithmetic

Example of data: (no,no,yes) (two,two,four) (send,more,money) (cross,road,danger) (donald,gerarld,robert)

Examples of Execution:
  python3 CryptoPuzzle.py
  python3 CryptoPuzzle.py -data=[send,more,money]
  python3 CryptoPuzzle.py -data=[send,more,money] -variant=carry
"""

from pycsp3 import *

word1, word2, word3 = words = [w.lower() for w in data] if data else ("no", "no", "yes")
n = len(word1)
assert len(word2) == n and len(word3) in {n, n + 1}

# x[i] is the value assigned to the ith letter (if present) of the alphabet
x = VarArray(size=26, dom=lambda i: range(10) if i in alphabet_positions("".join(words)) else None)

# auxiliary lists of variables associated with the three words
x1, x2, x3 = [[x[i] for i in reversed(alphabet_positions(word))] for word in words]

satisfy(
    # all letters must be assigned different values
    AllDifferent(x),

    # the most significant letter of each word cannot be equal to 0
    [x1[-1] != 0, x2[-1] != 0, x3[-1] != 0]
)

if not variant():
    satisfy(
        # ensuring the crypto-arithmetic sum
        Sum((x1[i] + x2[i]) * 10 ** i for i in range(n)) == Sum(x3[i] * 10 ** i for i in range(len(x3)))
    )

elif variant("carry"):
    # c[i] is the ith carry
    c = VarArray(size=n + 1, dom={0, 1})

    satisfy(
        # managing the least significant carry
        c[0] == 0,

        # managing the most significant carry
        c[n] == (0 if len(x3) == n else x3[n]),  # NB: the parentheses are required

        # managing remainders
        [(c[i] + x1[i] + x2[i]) % 10 == x3[i] for i in range(n)],

        # managing quotients
        [(c[i] + x1[i] + x2[i]) // 10 == c[i + 1] for i in range(n)]
    )



"""
See "Solving the Wastewater Treatment Plant Problem with SMT", by Miquel Bofill, Víctor Muñoz, Javier Murillo. CoRR abs/1609.05367 (2016)

Examples of Execution:
  python3 Wwtpp.py -data=Wwtpp_ex04400.json
  python3 Wwtpp.py -data=Wwtpp_ex04400.json -variant=short
"""

from pycsp3 import *

nIndustries, nPeriods, plantCapacity, tankFlow, tankCapacity, sd, spans = data  # sd for schedule flow of discharge


def table_compatibility(i, j):
    if j in {nPeriods - 1, nPeriods} or (j - 1 == 0 and sd[i][0] == 0):
        return {(0, 0)}
    tf, tc = tankFlow[i], tankCapacity[i]
    tbl = {(0, ANY)} | {(tf, v) for v in range(tf, tc + 1)} | {(v, v) for v in range(1, min(tf, tc + 1))}
    return tbl if variant("short") else to_ordinary_table(tbl, [tf + 1, tc + 1])


def table_spanning(i, start, stop):
    return {tuple([0] * (stop - start)), tuple(sd[i][start:stop])}


# b[i][j] is the flow stored in buffer i at the end of period j
b = VarArray(size=[nIndustries, nPeriods], dom=lambda i, j: {0} if (j == 0 and sd[i][0] == 0) or j == nPeriods - 1 else range(tankCapacity[i] + 1))

# d[i][j] is the flow discharged from buffer (or industry) i during time period j
d = VarArray(size=[nIndustries, nPeriods], dom=lambda i, j: {0} if j == 0 else range(tankFlow[i] + 1))

# c[i][j] is the actual capacity requirement of industry i during time period j
c = VarArray(size=[nIndustries, nPeriods], dom=lambda i, j: {0, sd[i][j]} if sd[i][j] != 0 else None)

satisfy(
    # not exceeding the Wastewater Treatment Plant
    [Sum(c[:, j] + d[:, j]) <= plantCapacity for j in range(nPeriods)],

    # managing scheduled discharge flows at period 0
    [Sum(b[i][0], c[i][0]) == sd[i][0] for i in range(nIndustries) if sd[i][0] != 0],

    # managing scheduled discharge flows at all periods except 0
    [[b[i][j], b[i][j - 1], d[i][j], c[i][j]] * [1, -1, 1, 1 if c[i][j] else None] == sd[i][j] for i in range(nIndustries) for j in range(1, nPeriods)],

    # ensuring compatibility between stored and discharge flows
    [(d[i][j], b[i][j - 1]) in table_compatibility(i, j) for i in range(nIndustries) for j in range(1, nPeriods)],

    # spanning constraints
    [c[i][start:stop] in table_spanning(i, start, stop) for (i, start, stop) in spans]
)


# Note that:

# a) when managing scheduled discharge flows, we have a list with four cells for variables and integers.
# However, when there is the special value None (at the fourth position in both lists), the two lists will be automatically reduced to three cells.



from pycsp3 import *
from pycsp3.solvers.abscon import AbsConProcess, AbsconPy4J

x = VarArray(size=[10, 10], dom=range(150))
y = VarArray(size=[10, 10], dom=range(150))
z = VarArray(size=10, dom=range(150))

satisfy(
    min(abs(x[0][0]), abs(y[0][0])),
    max(abs(x[0][0]), min(y[0], z[1]))
)

print("Compile:\n")
instance = compile()

print("\nStatic solving:\n")
solution = AbsConProcess().solve(instance)
print("solution:", solution)

print("\nPy4j solving:\n")
solver = AbsconPy4J()
solver.loadXCSP3(instance)

print("in progress")



"""
See Problem 008 on CSPLib

Examples of Execution:
  python3 VesselLoading.py -data=VesselLoading-inst2.json
"""

from pycsp3 import *

deckWidth, deckHeight, containers, separations = data
nContainers = len(containers)


def sep_pairs():
    groups = [[i for i, container in enumerate(containers) if container.type == k] for k in range(max(container.type for container in containers) + 1)]
    return [(i, j, distance) for (type1, type2, distance) in separations for i in groups[type1] for j in groups[type2]]


# x[i] is the x-coordinate of the ith container
x = VarArray(size=nContainers, dom=range(deckWidth))

# y[i] is the y-coordinate of the ith container
y = VarArray(size=nContainers, dom=range(deckHeight))

# w[i] is the width of the ith container
w = VarArray(size=nContainers, dom=lambda i: {containers[i].width, containers[i].height})

# h[i] is the height of the ith container
h = VarArray(size=nContainers, dom=lambda i: {containers[i].width, containers[i].height})

# r[i] is 1 iff the ith container is rotated by 90 degrees
r = VarArray(size=nContainers, dom={0, 1})

satisfy(
    # horizontal control
    [x[i] + w[i] <= deckWidth for i in range(nContainers)],

    # vertical control
    [y[i] + h[i] <= deckHeight for i in range(nContainers)],

    # managing rotation
    [(r[i], w[i], h[i]) in {(0, width, height), (1, height, width)} for i, (width, height, _) in enumerate(containers)],

    # no overlapping between containers
    NoOverlap(origins=[(x[i], y[i]) for i in range(nContainers)], lengths=[(w[i], h[i]) for i in range(nContainers)]),

    # respecting separations between containers according to their types
    [(x[i] + w[i] + sep <= x[j]) | (x[j] + w[j] + sep <= x[i]) | (y[i] + h[i] + sep <= y[j]) | (y[j] + h[j] + sep <= y[i]) for (i, j, sep) in sep_pairs()]
)



from pycsp3 import *


n, p = data  # n is the board size -- p is the number of pions

x = VarArray(size=p, dom=range(n * n))

satisfy(
    (dist(x[i], x[j]) < p - 1) & (x[i] != x[j]) for i in range(p) for j in range(i + 1, p)
)



"""
Problem 023 on CSPLib

Examples of Execution:
  python3 MagicHexagon.py
  python3 MagicHexagon.py -data=[4,10]
"""

from pycsp3 import *

n, s = data or (4, 3)
gap = 3 * n * n - 3 * n + 1
sum_gap = sum(range(s, s + gap))
assert sum_gap % (2 * n - 1) == 0, "No magic hexagon for order=" + str(n) + " and start=" + str(s)
magic = sum_gap // (2 * n - 1)
d = n + n - 1  # longest diameter


def scope(i, right):
    v1 = max(0, d // 2 - i) if right else max(0, i - d // 2)
    v2 = d // 2 - v1
    return [x[j + v1][i - max(0, v2 - j if right else j - v2)] for j in range(d - abs(d // 2 - i))]


# x represents the hexagon; on row x[i], only the first n - |n/2 - i| cells are useful (here, n = 'd').
x = VarArray(size=[d, d], dom=lambda i, j: range(s, s + gap) if j < d - abs(d // 2 - i) else None)

satisfy(
    AllDifferent(x),

    # all rows sum to the magic value
    [Sum(x[i]) == magic for i in range(d)],

    # all right-sloping diagonals sum to the magic value
    [Sum(scope(i, True)) == magic for i in range(d)],

    # all left-sloping diagonals sum to the magic value
    [Sum(scope(i, False)) == magic for i in range(d)],

    # tag(symmetry-breaking)
    [
        x[0][0] < x[0][n - 1],
        x[0][0] < x[n - 1][d - 1],
        x[0][0] < x[d - 1][n - 1],
        x[0][0] < x[d - 1][0],
        x[0][0] < x[n - 1][0],
        x[0][n - 1] < x[n - 1][0]
    ]
)



"""
Problem 016 on CSPLib

Execution:
  python3 TrafficLights.py
"""

from pycsp3 import *

R, RY, G, Y = "red", "red-yellow", "green", "yellow"

table = {(R, R, G, G), (RY, R, Y, R), (G, G, R, R), (Y, R, RY, R)}

# v[i] is the color for the ith vehicle traffic light
v = VarArray(size=4, dom={R, RY, G, Y})

# p[i] is the color for the ith pedestrian traffic light
p = VarArray(size=4, dom={R, G})

satisfy(
    (v[i], p[i], v[(i + 1) % 4], p[(i + 1) % 4]) in table for i in range(4)
)



"""
Problem proposed by A. Ollagnier and J.-G. Fages for the 2013 Minizinc Competition.

Examples of Execution:
  python3 Mario.py -data=Mario_easy-2.json
  python3 Mario.py -data=Mario_easy-2.json -variant=table
"""

from pycsp3 import *

marioHouse, luigiHouse, fuelLimit, houses = data
fuels = [house.fuelConsumption for house in houses]  # using cp_array is not necessary since intern arrays have the right type (for the constraint Element)
nHouses = len(houses)

# s[i] is the house succeeding to the ith house (itself if not part of the route)
s = VarArray(size=nHouses, dom=range(nHouses))

# f[i] is the fuel consumed at each step (from house i to its successor)
f = VarArray(size=nHouses, dom=lambda i: fuels[i])

if not variant():
    satisfy(
        # fuel consumption at each step
        fuels[i][s[i]] == f[i] for i in range(nHouses)
    )

elif variant("table"):
    satisfy(
        # fuel consumption at each step
        (s[i], f[i]) in {(j, fuel) for (j, fuel) in enumerate(house.fuelConsumption)} for i, house in enumerate(houses)
    )

satisfy(
    # we cannot consume more than the available fuel
    Sum(f) <= fuelLimit,

    # Mario must make a tour (not necessarily complete)
    Circuit(s),

    # Mario's house succeeds to Luigi's house
    s[luigiHouse] == marioHouse
)

maximize(
    # maximizing collected gold
    Sum((s[i] != i) * houses[i].gold for i in range(nHouses) if i not in {marioHouse, luigiHouse})
)


# [(s[i], f[i]) in [(j, houses[i].fuelConsumption[j]) for j in range(len(houses[i].fuelConsumption))] for i in range(nHouses)],

# [(s[i], f[i]) in [(j, fuel) for j, fuel in enumerate(houses[i].fuelConsumption)] for i in range(nHouses)],

# g[i] is the gold earned at house i
# g = VarArray(size=nHouses, dom=lambda i: {0, houses[i].gold})
# gold earned at each house
# maximize(
#    maximizing collected gold
#    Sum(g)
# )



"""
Problem 050 on CSPLib

Examples of Execution:
  python3 DiamondFree.py
  python3 DiamondFree.py -data=10
"""

from pycsp3 import *

n = data or 8

# x is the adjacency matrix
x = VarArray(size=[n, n], dom=lambda i, j: {0, 1} if i != j else {0})

# y[i] is the degree of the ith node
y = VarArray(size=n, dom={i for i in range(1, n) if i % 3 == 0})

# s is the sum of all degrees
s = Var(dom={i for i in range(n, n * (n - 1) + 1) if i % 12 == 0})

satisfy(
    # ensuring the absence of diamond in the graph
    [Sum(x[i][j], x[i][k], x[i][l], x[j][k], x[j][l], x[k][l]) <= 4 for i, j, k, l in combinations(range(n), 4)],

    # ensuring that the graph is undirected (symmetric)
    [x[i][j] == x[j][i] for i, j in combinations(range(n), 2)],

    # computing node degrees
    [Sum(x[i]) == y[i] for i in range(n)],

    # computing the sum of node degrees
    Sum(y) == s,

    # tag(symmetry-breaking)
    [Decreasing(y), LexIncreasing(x)]
)



from pycsp3 import *

t = cp_array([[1, 2, 3], [4, 5, 6]])
a = Var(0, 1)
b = Var(0, 1, 2)
r = Var(range(1, 7))

x = VarArray(size=[2, 2], dom=range(150))

y = Var(dom=range(5))
z = Var(dom=range(5))


satisfy(
    t[a][b] == r,
    a + b + 1 == r
    # ,x[y][z] == 1
)



from pycsp3.problems.data.parsing import *

data["n"] = number_in(line())
data["m"] = number_in(next_line())

nPairs = number_in(next_line())
p1 = list(zip(numbers_in(next_line()), numbers_in(next_line())))
p2 = list(zip(numbers_in(next_line()), numbers_in(next_line())))
assert nPairs == len(p1) == len(p2)
data["points1"] = p1
data["points2"] = p2  #[[p1[i], p2[i]] for i in range(len(p1))]




"""
Problem 019 on CSPLib

Examples of Execution:
  python3 MagicSequence.py
  python3 MagicSequence.py -data=10
"""

from pycsp3 import *

n = data or 8

# x[i] is the ith value of the sequence
x = VarArray(size=n, dom=range(n))

satisfy(
    # each value i occurs exactly x[i] times in the sequence
    Cardinality(x, occurrences={i: x[i] for i in range(n)}),

    # tag(redundant-constraints)
    [Sum(x) == n, Sum((i - 1) * x[i] for i in range(n)) == 0]
)

# Note that

# a) Sum((i - 1) * x[i] for i in range(n)) == 0
# could be equivalently written x * range(-1, n - 1) == 0
# but range(-1, n - 1) * x == 0 is currently not possible (requires 'cursing' * for range objects)



"""
Betty, Chris, Donald, Fred, Gary, Mary, and Paul want to align in one row for taking a photo.
Some of them have preferences next to whom they want to stand:
 - Betty wants to stand next to Gary and Mary.
 - Chris wants to stand next to Betty and Gary.
 - Fred wants to stand next to Mary and Donald.
 - Paul wants to stand next to Fred and Donald.

Examples of Execution:
  python3 Photo.py
  python3 Photo.py -variant=aux
"""

from pycsp3 import *

# friends[i] is the position (in a row) of the ith friend
betty, chris, donald, fred, gary, mary, paul = friends = VarArray(size=7, dom=range(7))

preferences = [(betty, gary), (betty, mary), (chris, betty), (chris, gary), (fred, mary), (fred, donald), (paul, fred), (paul, donald)]

if not variant():
    satisfy(
        # all friends are at a different position
        AllDifferent(friends)
    )

    minimize(
        # minimizing the overall dissatisfaction
        Sum(abs(f1 - f2) != 1 for (f1, f2) in preferences)
    )

elif variant("aux"):
    # costs[i] is the cost of not respecting the ith preference
    costs = VarArray(size=len(preferences), dom={0, 1})

    table = {(i, j, 0 if abs(i - j) == 1 else 1) for i in range(7) for j in range(7) if i != j}

    satisfy(
        # all friends are at a different position
        AllDifferent(friends),

        # determining which preferences are not satisfied
        [(f1, f2, costs[i]) in table for i, (f1, f2) in enumerate(preferences)]
    )

    minimize(
        # minimizing the overall dissatisfaction
        Sum(costs)
    )



"""
The queens graph is a graph with n*n nodes corresponding to the squares of a chess-board.
There is an edge between nodes iff they are on the same row, column, or diagonal, i.e. if two queens on those squares would attack each other.
The coloring problem is to color the queens graph with n colors.
See Tom Kelsey, Steve Linton, Colva M. Roney-Dougal: New Developments in Symmetry Breaking in Search Using Computational Group Theory. AISC 2004: 199-210.

Examples of Execution:
  python3 ColouredQueens.py
  python3 ColouredQueens.py -data=10
"""

from pycsp3 import *

n = data or 8

# x[i][j] is the color at row i and column j
x = VarArray(size=[n, n], dom=range(n))

satisfy(
    # different colors on rows and columns
    AllDifferent(x, matrix=True),

    # different colors on downward diagonals
    [AllDifferent(dgn) for dgn in diagonals_down(x)],

    # different colors on upward diagonals
    [AllDifferent(dgn) for dgn in diagonals_up(x)]
)



"""
"""
from pycsp3 import *

height, width = data.height, data.width
bugs, lengths = data.bugs, data.bugTypeLengths
nBugs, nTypes = len(bugs), len(lengths)
groups = [[i for i in range(nBugs) if data.bugs[i].type == k] for k in range(nTypes)]  # groups[k] is the group of bugs of type k

# x[i][j] is the index of a bug (or -1)
x = VarArray(size=[height, width], dom=range(-1, nBugs))

# n1[i] is the number of cells with value i
n1 = VarArray(size=nBugs, dom=range(height * width + 1))

# n2[i] is the number of cells with with value i and a bug on it
n2 = VarArray(size=nBugs, dom=range(nBugs + 1))


def neighbors(i, j, k, l):
    return i == k and j in {l - 1, l + 1} or j == l and i in {k - 1, k + 1}


satisfy(
    # each bug cell can take either its bug index for value or the one of another bug of the same type
    [(x[bug.row][bug.col]) in groups[bug.type] for bug in bugs],

    # a bug cell value is smaller or equal to the bug index  tag(symmetry-breaking)
    [x[bug.row][bug.col] <= i for i, bug in enumerate(bugs)],

    [n1[group[j]] <= (len(group) - j) * lengths[i] for i, group in enumerate(groups) for j in range(len(group))],

    # computing values of n1
    Cardinality(x, occurrences={i: n1[i] for i in range(nBugs)}),

    # computing values of n2
    Cardinality([x[bug.row][bug.col] for bug in bugs], occurrences={i: n2[i] for i in range(nBugs)}),

    # linking variables of n1 and n2
    [n2[i] * lengths[bugs[i].type] == n1[i] for i in range(nBugs)],

    # if two cells have the same value, then all cells in the rectangle they delimit must take this value
    [imply(x[i][j] == x[k][l], x[i][j] == x[m][n]) for (i, j, k, l, m, n) in product(range(height), range(width), repeat=3) if
     i <= m <= k and not neighbors(i, j, k, l) and (i != k or j < l) and min(j, l) <= n <= max(j, l) and (m, n) != (i, j) and (m, n) != (k, l)]
)

maximize(
    # maximizing the number of times a bug is in a rectangle whose index is different from the one of the bug
    Sum(n2[i] == 0 for i in range(nBugs))
)



from pycsp3.problems.data.parsing import *

next_line()
data["height"] = next_int()
data["width"] = next_int()
next_line()
nBugTypes = next_int()
next_line()
data["bugTypeLengths"] = [next_int() for _ in range(nBugTypes)]
next_line()
nBugs = next_int()
next_line()
data["bugs"] = [OrderedDict([("row", next_int()), ("col", next_int()), ("type", next_int())]) for _ in range(nBugs)]
# data.bugTypeLengths = bugTypesLength  # [DataDict({"length": bugTypesLength[i], "cells": [j for j in range(nBugs) if data.bugs[j].type == i]}) for i in range(nBugTypes)]



from pycsp3.problems.tests.tester import Tester, run

NAME = "g3_pattern"

diff = (Tester(NAME)
        .add("Dominoes", data="Dominoes_grid1.json")  # ok, unary constraints managed differently
        .add("Fastfood", data="Fastfood_ff01.json")
        .add("Fastfood", data="Fastfood_ff01.json", variant="table")
        .add("Fastfood", data="Fastfood_example.dzn", prs_py="Fastfood_ParserZ", prs_jv="Fastfood_ParserZ")
        .add("Kakuro", data="Kakuro_easy-000.json")
        .add("Lightup", data="Lightup_example.txt", prs_py="Lightup_Parser", prs_jv="Lightup_Parser")  # ok, JvCSP more compact because reordering vars in count
        .add("MarketSplit", data="MarketSplit_04.json")
        .add("MultiKnapsack", data="MultiKnapsack_example.txt", prs_py="MultiKnapsack_Parser", prs_jv="MultiKnapsack_Parser")
        .add("OpenStacks", data="OpenStacks_example.dzn", prs_py="OpenStacks_ParserZ", prs_jv="OpenStacks_ParserZ", variant="m1")
        .add("OpenStacks", data="OpenStacks_example.dzn", prs_py="OpenStacks_ParserZ", prs_jv="OpenStacks_ParserZ", variant="m2")
        .add("ProgressiveParty", data="ProgressiveParty_example.txt", prs_py="ProgressiveParty_Parser", prs_jv="ProgressiveParty_Parser", special=True)
        # ok, group of alldiffExcept in PyCSp contrary to JvCSP
        # .add("BinPacking", data="BinPacking_n1c1w4a.json", variant="mdd", special=True)
        )

same = (Tester(NAME)
        .add("Auction", data="Auction_example.json")
        .add("Auction", data="Auction_example.txt", prs_py="Auction_Parser", prs_jv="Auction_Parser")
        .add("BinPacking", data="BinPacking_n1c1w4a.json")
        .add("BinPacking", data="BinPacking_n1c1w4a.json", variant="table")
        .add("BinPacking", data="BinPacking_example.txt", prs_py="BinPacking_Parser", prs_jv="BinPacking_Parser")
        .add("Blackhole", data="Blackhole.json")
        .add("Bugs", data="Bugs_example.json")
        .add("Bugs", data="Bugs_example.txt", prs_py="Bugs_Parser", prs_jv="Bugs_Parser")
        .add("BusScheduling", data="BusScheduling_c1a.json")
        .add("BusScheduling", data="BusScheduling_r1.json")
        .add("BusScheduling", data="BusScheduling_t1.json")
        .add("Coloring", data="Coloring_rand1.json")
        .add("Coloring", data="Coloring_rand1.json", variant="csp")
        .add("Cutstock", data="Cutstock_small.json")
        .add("Eternity", data="Eternity_07x05.json")
        .add("Eternity", data="Eternity_example.txt", prs_py="Eternity_Parser", prs_jv="Eternity_Parser")
        .add("Futoshiki", data="Futoshiki_futo3_0.json")
        .add("GraphColoring", data="GraphColoring_1-fullins-3.json")
        .add("GraphColoring", data="GraphColoring_1-fullins-3.json", variant="sum")
        .add("GraphColoring", data="GraphColoring_qwhdec-o18-h120-1.json")
        .add("GraphColoring", data="GraphColoring_qwhdec-o18-h120-1.json", variant="sum")
        .add("GraphMaxAcyclic", data="GraphMaxAcyclic_example.txt", prs_py="GraphMaxAcyclic_Parser", prs_jv="GraphMaxAcyclic_Parser")
        .add("GraphMaxAcyclic", data="GraphMaxAcyclic_example.txt", prs_py="GraphMaxAcyclic_Parser", prs_jv="GraphMaxAcyclic_Parser", variant="cnt")
        .add("HCPizza", data="HCPizza_tiny.txt", prs_py="HCPizza_Parser", prs_jv="HCPizza_Parser")
        .add("HCPizza", data="HCPizza_small.txt", prs_py="HCPizza_Parser", prs_jv="HCPizza_Parser")
        .add("Kakuro", data="Kakuro_easy-000.json", variant="table")
        .add("Knapsack", data="Knapsack_20-50-00.json")
        .add("LatinSquare", data="LatinSquare_qwh-o030-h320.json")
        .add("LatinSquareB", data="LatinSquare2_7-2-0.json")
        .add("LinearArrangement", data="LinearArrangement_04.json")
        .add("MagicSquare", data="MagicSquare_example0.txt", prs_py="MagicSquare_Parser", prs_jv="MagicSquare_Parser")
        .add("MagicSquare", data="MagicSquare_example1.txt", prs_py="MagicSquare_Parser", prs_jv="MagicSquare_Parser")
        .add("Mario", data="Mario_easy-2.json")
        .add("Mario", data="Mario_easy-2.json", variant="table")
        .add("Nonogram", data="Nonogram_example.txt", prs_py="Nonogram_Parser", prs_jv="Nonogram_Parser")
        .add("Nonogram", data="Nonogram_example.txt", prs_py="Nonogram_Parser", prs_jv="Nonogram_Parser", variant="table")
        .add("PrizeCollecting", data="PrizeCollecting_example.dzn", prs_py="PrizeCollecting_ParserZ", prs_jv="PrizeCollecting_ParserZ", special=True)
        .add("PseudoBoolean", data="PseudoBoolean_example.opb", prs_py="PseudoBoolean_Parser", prs_jv="PseudoBoolean_Parser")
        .add("RectPacking", data="RectPacking_perfect-001.json", special=True)
        .add("RoomMate", data="RoomMate_sr0006.json")
        .add("Sat", data="Sat_flat30-16.json", variant="sum")
        .add("Sat", data="Sat_flat30-16.json", variant="dual")
        .add("Sat", data="Sat_flat30-16.json", variant="clause")
        .add("Shikaku", data="Shikaku_grid1.json")
        .add("Sonet", data="Sonet_sonet1.json")
        .add("Sonet", data="Sonet_sonet3-4.json")
        .add("StripPacking", data="StripPacking_C1P1.json")
        .add("Subisomorphism", data="Subisomorphism_A-01.json")
        .add("Sudoku", data="Sudoku_s13a.json")
        .add("Sudoku", data="Sudoku_s13a.json", variant="table")
        .add("Sudoku", data="Sudoku_example.txt", prs_py="Sudoku_Parser", prs_jv="Sudoku_Parser")
        .add("TravelingSalesman", data="TravelingSalesman_10-20-0.json")
        .add("TravelingSalesman", data="TravelingSalesman_10-20-0.json", variant="table")
        .add("VesselLoading", data="VesselLoading-inst1.json")
        )

xcsp = (Tester(NAME)
        .add("Amaze", data="Amaze_simple.json")  #
        .add("Amaze", data="Amaze_2012-03-07.dzn", prs_py="Amaze_ParserZ.py", prs_jv="")  #
        .add("Areas", data="Areas-3-3-3.json")  # 7 solutions
        .add("Auction", data="Auction_example.json")  # optimum 54
        .add("Auction", data="Auction_example.txt", prs_py="Auction_Parser", prs_jv="Auction_Parser")  # optimum 54
        .add("BinPacking", data="BinPacking_n1c1w4a.json")  # optimum 5
        .add("BinPacking", data="BinPacking_n1c1w4a.json", variant="table")  # optimum 5
        .add("BinPacking", data="BinPacking_example.txt", prs_py="BinPacking_Parser", prs_jv="BinPacking_Parser")  # optimum 5
        .add("Blackhole", data="Blackhole.json", options_py="-recognizeSlides")  # 47232 solutions
        .add("Bugs", data="Bugs_example.json")  # optimum 5
        .add("Bugs", data="Bugs_example.txt", prs_py="Bugs_Parser", prs_jv="Bugs_Parser")  # optimum 5
        .add("BusScheduling", data="BusScheduling_c1a.json")
        .add("BusScheduling", data="BusScheduling_r1.json")
        .add("BusScheduling", data="BusScheduling_t1.json")  # optimum 7
        .add("Coloring", data="Coloring_rand1.json")  # optimum 2
        .add("Coloring", data="Coloring_rand1.json", variant="csp")  # sat
        .add("Cutstock", data="Cutstock_small.json")  # optimum 4
        .add("Dominoes", data="Dominoes_grid1.json")  # 128 solutions
        .add("Eternity", data="Eternity_07x05.json")  # 32 solutions
        .add("Eternity", data="Eternity_example.txt", prs_py="Eternity_Parser", prs_jv="Eternity_Parser")  # 32 solutions
        .add("Fastfood", data="Fastfood_ff01.json")  # optimum 3050
        .add("Fastfood", data="Fastfood_ff01.json", variant="table") # optimum 3050
        .add("Fastfood", data="Fastfood_example.dzn", prs_py="Fastfood_ParserZ", prs_jv="Fastfood_ParserZ")  # optimum 3050
        .add("Futoshiki", data="Futoshiki_futo3_0.json")  # 1 solution
        .add("GraphColoring", data="GraphColoring_1-fullins-3.json")  # optimum 3
        .add("GraphColoring", data="GraphColoring_1-fullins-3.json", variant="sum")  # optimum 24
        .add("GraphColoring", data="GraphColoring_qwhdec-o18-h120-1.json")  # optimum 17
        .add("GraphColoring", data="GraphColoring_qwhdec-o18-h120-1.json", variant="sum")  # optimum 2754
        .add("GraphMaxAcyclic", data="GraphMaxAcyclic_example.json")  # optimum 44
        .add("GraphMaxAcyclic", data="GraphMaxAcyclic_example.txt", prs_py="GraphMaxAcyclic_Parser", prs_jv="GraphMaxAcyclic_Parser")  # optimum 44
        .add("GraphMaxAcyclic", data="GraphMaxAcyclic_example.txt", prs_py="GraphMaxAcyclic_Parser", prs_jv="GraphMaxAcyclic_Parser", variant="cnt")  # optimum 44
        .add("HCPizza", data="HCPizza_tiny.json")  # optimum 15
        .add("HCPizza", data="HCPizza_tiny.txt", prs_py="HCPizza_Parser", prs_jv="HCPizza_Parser")  # optimum 15
        .add("HCPizza", data="HCPizza_small.txt", prs_py="HCPizza_Parser", prs_jv="HCPizza_Parser")
        .add("Kakuro", data="Kakuro_easy-000.json")  # 1 solution
        .add("Kakuro", data="Kakuro_easy-000.json", variant="table")  # 1 solution
        .add("Knapsack", data="Knapsack_20-50-00.json")  # optimum 583
        .add("LatinSquare", data="LatinSquare_qwh-o030-h320.json")
        .add("LatinSquare2", data="LatinSquare2_7-2-0.json")  # 480 solutions
        .add("Lightup", data="Lightup_example.json")  # 1 solution
        .add("Lightup", data="Lightup_example.txt", prs_py="Lightup_Parser", prs_jv="Lightup_Parser")  # 1 solution
        .add("LinearArrangement", data="MinLA01.json")
        .add("MagicSquare", data="[4,None]")  # 7040 solutions
        .add("MagicSquare", data="MagicSquare_example1.txt", prs_py="MagicSquare_Parser", prs_jv="MagicSquare_Parser")
        .add("Mario", data="Mario_easy-2.json")  # optimum 628
        .add("Mario", data="Mario_easy-2.json", variant="table")  # optimum 628
        .add("MarketSplit", data="MarketSplit_04.json")  # 1 solution
        .add("MultiKnapsack", data="MultiKnapsack_example.txt", prs_py="MultiKnapsack_Parser", prs_jv="MultiKnapsack_Parser")
        .add("Nonogram", data="Nonogram_example.txt", prs_py="Nonogram_Parser", prs_jv="Nonogram_Parser")  # 1 solution
        .add("Nonogram", data="Nonogram_example.txt", prs_py="Nonogram_Parser", prs_jv="Nonogram_Parser", variant="table")  # 1 solution
        .add("OpenStacks", data="OpenStacks_example.dzn", prs_py="OpenStacks_ParserZ", prs_jv="OpenStacks_ParserZ", variant="m1")  # optimum 45
        .add("OpenStacks", data="OpenStacks_example.dzn", prs_py="OpenStacks_ParserZ", prs_jv="OpenStacks_ParserZ", variant="m2")  # optimum 45
        .add("PrizeCollecting", data="PrizeCollecting_example.dzn", prs_py="PrizeCollecting_ParserZ", prs_jv="PrizeCollecting_ParserZ", special=True)
        .add("ProgressiveParty", data="ProgressiveParty_example.txt", prs_py="ProgressiveParty_Parser", prs_jv="ProgressiveParty_Parser", special=True) # optimum 5
        .add("PseudoBoolean", data="PseudoBoolean_example.opb", prs_py="PseudoBoolean_Parser", prs_jv="PseudoBoolean_Parser")  # optimum 20
        .add("RectPacking", data="RectPacking_perfect-001.json", special=True)
        .add("RoomMate", data="RoomMate_sr0006.json")  # 2 solutions
        .add("Sat", data="Sat_flat30-16.json", variant="clause")  # 1482 solutions
        .add("Sat", data="Sat_flat30-16.json", variant="sum")  # 1482 solutions
        .add("Sat", data="Sat_flat30-16.json", variant="dual")  # 1482 solutions
        .add("Shikaku", data="Shikaku_grid1.json")  # 1 solution
        .add("Sonet", data="Sonet_sonet1.json")  # optimum 8
        .add("Sonet", data="Sonet_sonet3-4.json")  # optimum 12
        .add("StripPacking", data="StripPacking_C1P1.json")
        .add("Subisomorphism", data="Subisomorphism_A-01.json")  # 1 solution
        .add("Sudoku", data="[9,None]")  # None can be replaced by null (but not by an empty string in this context)
        .add("Sudoku", data="Sudoku_s13a.json")  # 1 solution
        .add("Sudoku", data="Sudoku_s13a.json", variant="table")  # 1 solution
        .add("Sudoku", data="Sudoku_example.txt", prs_py="Sudoku_Parser", prs_jv="Sudoku_Parser")  # 1 solution
        .add("TravelingSalesman", data="TravelingSalesman_10-20-0.json")  # optimum 47
        .add("TravelingSalesman", data="TravelingSalesman_10-20-0.json", variant="table")  # optimum 47
        .add("VesselLoading", data="VesselLoading-inst1.json")  # 8 solutions
        .add("VesselLoading", data="VesselLoading-inst2.json")
        )

run(xcsp,diff, same)



"""
Problem 003 on CSPLib

Examples of Execution:
  python3 QuasiGroup.py
  python3 QuasiGroup.py -data=8 -variant=base-v3
  python3 QuasiGroup.py -data=5 -variant=base-v4
  python3 QuasiGroup.py -data=8 -variant=base-v5
  python3 QuasiGroup.py -data=8 -variant=base-v6
  python3 QuasiGroup.py -data=9 -variant=base-v7
  python3 QuasiGroup.py -data=8 -variant=aux-v3
  python3 QuasiGroup.py -data=5 -variant=aux-v4
  python3 QuasiGroup.py -data=8 -variant=aux-v5
  python3 QuasiGroup.py -data=9 -variant=aux-v7
 """

from pycsp3 import *

n = data or 8

# x[i][j] is the value at row i and column j of the quasi-group
x = VarArray(size=[n, n], dom=range(n))

satisfy(
    # ensuring a Latin square
    AllDifferent(x, matrix=True),

    # ensuring idempotence  tag(idempotence)
    [x[i][i] == i for i in range(n)]
)

if variant("base"):
    if subvariant("v3"):
        satisfy(
            x[x[i][j], x[j][i]] == i for i in range(n) for j in range(n)
        )
    elif subvariant("v4"):
        satisfy(
            x[x[j, i], x[i, j]] == i for i in range(n) for j in range(n)
        )
    elif subvariant("v5"):
        satisfy(
            x[x[x[j, i], j], j] == i for i in range(n) for j in range(n)
        )
    elif subvariant("v6"):
        satisfy(
            x[x[i, j], j] == x[i, x[i, j]] for i in range(n) for j in range(n)
        )
    elif subvariant("v7"):
        satisfy(
            x[x[j, i], j] == x[i, x[j, i]] for i in range(n) for j in range(n)
        )
elif variant("aux"):
    if subvariant("v3"):
        y = VarArray(size=[n, n], dom=range(n * n))

        satisfy(
            [x[y[i][j]] == i for i in range(n) for j in range(n) if i != j],

            [y[i][j] == x[i][j] * n + x[j][i] for i in range(n) for j in range(n) if i != j]
        )
    elif subvariant("v4"):
        y = VarArray(size=[n, n], dom=range(n * n))

        satisfy(
            [x[y[i][j]] == i for i in range(n) for j in range(n) if i != j],

            [y[i][j] == x[j][i] * n + x[i][j] for i in range(n) for j in range(n) if i != j]
        )
    elif subvariant("v5"):
        y = VarArray(size=[n, n], dom=range(n))

        satisfy(
            [x[:, i][x[i][j]] == y[i][j] for i in range(n) for j in range(n) if i != j],

            [x[:, i][y[i][j]] == j for i in range(n) for j in range(n) if i != j]
        )
    elif subvariant("v7"):
        y = VarArray(size=[n, n], dom=range(n))

        satisfy(
            (col[x[j][i]] == y[i][j], x[i][x[j][i]] == y[i][j]) for i in range(n) for j, col in enumerate(columns(x)) if i != j
        )



from pycsp3.problems.data.parsing import *

data["nCourses"] = number_in(next_line())
data["nPeriods"] = number_in(next_line())
data["minCredits"] = number_in(next_line())
data["maxCredits"] = number_in(next_line())
data["minCourses"] = number_in(next_line())
data["maxCourses"] = number_in(next_line())
data["credits"] = numbers_in(next_line())

assert data["nCourses"] == len(data["credits"])

data["prerequisites"] = [[int(v) - 1 for v in re.split(r'constraint prerequisite\(|,|\);', line) if len(v) > 0] for line in remaining_lines(skip_curr=True)]



"""
See, e.g., "AllDifferent-based filtering for subgraph isomorphism"  by Christine Solnon, Artificial Intelligence, 174(12-13): 850-864 (2010)

Example of Execution:
  python3 Subisomorphism.py -data=Subisomorphism_A-01.json
"""

from pycsp3 import *

n, m, p_edges, t_edges = data


def structures():
    p_degrees = [len([edge for edge in p_edges if i in edge]) for i in range(n)]
    t_degrees = [len([edge for edge in t_edges if i in edge]) for i in range(m)]
    both_way_table = {(i, j) for (i, j) in t_edges} | {(j, i) for (i, j) in t_edges}
    degree_conflicts = [{j for j in range(m) if t_degrees[j] < p_degrees[i]} for i in range(n)]
    return [i for (i, j) in p_edges if i == j], [i for (i, j) in t_edges if i == j], both_way_table, degree_conflicts


p_loops, t_loops, table, degree_conflicts = structures()

# x[i] is the target node to which the ith pattern node is mapped
x = VarArray(size=n, dom=range(m))

satisfy(
    # ensuring injectivity
    AllDifferent(x),

    # preserving edges
    [(x[i], x[j]) in table for (i, j) in p_edges],

    # being careful of self-loops
    [x[i] in t_loops for i in p_loops],

    # tag(redundant-constraints)
    [x[i] not in conflicts for i, conflicts in enumerate(degree_conflicts)]
)



from pycsp3.problems.data.parsing import *

nRows, nCols = next_int(), next_int()
data["rowPatterns"] = [[next_int() for _ in range(next_int())] for _ in range(nRows)]
data["colPatterns"] = [[next_int() for _ in range(next_int())] for _ in range(nRows)]



import os

from pycsp3.solvers.solver import SolverProcess, SolverPy4J, directory_of_solver, class_path_abscon


class AbsConProcess(SolverProcess):
    def __init__(self):
        c = class_path_abscon()
        super().__init__(name="AbsCon", command="java -cp " + c + " AbsCon")


class AbsconPy4J(SolverPy4J):
    def __init__(self):
        d = directory_of_solver("abscon")
        c = class_path_abscon()
        super().__init__(name="AbsCon", command="java -cp " + c + os.pathsep + d + "py4j0.10.8.1.jar" + os.pathsep + d + " AbsConPy4J")

# command="java -cp /usr/local/share/py4j/py4j0.10.8.1.jar:.:./pyAbsCon/ StackEntryPoint"



""""
See https://en.wikipedia.org/wiki/Sudoku
See, e.g., "Sudoku as a Constraint Problem" by Helmut Simonis

Example of Execution:
  python3 Sudoku.py
  python3 Sudoku.py -data=null
  python3 Sudoku.py -data=None
  python3 Sudoku.py -data=Sudoku9-test.json
"""

from pycsp3 import *

clues = data  # if not 0, clues[i][j] is a value imposed at row i and col j




# x[i][j] is the value in cell at row i and col j.
x = VarArray(size=[9, 9], dom=range(1, 10))

satisfy(
    # imposing distinct values on each row and each column
    AllDifferent(x, matrix=True),

    # imposing distinct values on each block  tag(blocks)
    [AllDifferent(x[i:i + 3, j:j + 3]) for i in [0, 3, 6] for j in [0, 3, 6]],

    # imposing clues  tag(clues)
    [x[i][j] == clues[i][j] for i in range(9) for j in range(9) if clues and clues[i][j] > 0]
)
